"""Huggingface transformers ModelClient integration."""

from typing import Any, Dict, Union, List, Optional, Sequence
import logging
from functools import lru_cache
import re
import warnings


from adalflow.core.model_client import ModelClient
from adalflow.core.types import GeneratorOutput, ModelType, Embedding, EmbedderOutput
from adalflow.core.functional import get_top_k_indices_scores

# optional import
from adalflow.utils.lazy_import import safe_import, OptionalPackages


transformers = safe_import(
    OptionalPackages.TRANSFORMERS.value[0], OptionalPackages.TRANSFORMERS.value[1]
)
torch = safe_import(OptionalPackages.TORCH.value[0], OptionalPackages.TORCH.value[1])

import torch

import torch.nn.functional as F
from torch import Tensor

from transformers import (
    AutoTokenizer,
    AutoModel,
    AutoModelForSequenceClassification,
)


log = logging.getLogger(__name__)


def average_pool(last_hidden_states: Tensor, attention_mask: Tensor) -> Tensor:
    last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)
    return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]


# TODO: provide a standard api for embedding and chat models used in local model SDKs
class TransformerEmbedder:
    """Local model SDK for transformers.


    There are two ways to run transformers:
    (1) model and then run model inference
    (2) Pipeline and then run pipeline inference

    This file demonstrates how to
    (1) create a torch model inference component:  TransformerEmbedder which equalize to OpenAI(), the SyncAPIClient
    (2) Convert this model inference component to LightRAG API client: TransformersClient

    The is now just an exmplary component that initialize a certain model from transformers and run inference on it.
    It is not tested on all transformer models yet. It might be necessary to write one for each model.

    References:
    - transformers: https://huggingface.co/docs/transformers/en/index
    - thenlper/gte-base model:https://huggingface.co/thenlper/gte-base
    """

    models: Dict[str, type] = {}

    def __init__(self, model_name: Optional[str] = "thenlper/gte-base"):
        super().__init__()

        if model_name is not None:
            self.init_model(model_name=model_name)

    @lru_cache(None)
    def init_model(self, model_name: str):
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name)
            self.model = AutoModel.from_pretrained(model_name)
            # register the model
            self.models[model_name] = self.model
            log.info(f"Done loading model {model_name}")

        except Exception as e:
            log.error(f"Error loading model {model_name}: {e}")
            raise e

    def infer_gte_base_embedding(
        self,
        input=Union[str, List[str]],
        tolist: bool = True,
    ):
        model = self.models.get("thenlper/gte-base", None)
        if model is None:
            # initialize the model
            self.init_model("thenlper/gte-base")

        if isinstance(input, str):
            input = [input]
        # Tokenize the input texts
        batch_dict = self.tokenizer(
            input, max_length=512, padding=True, truncation=True, return_tensors="pt"
        )
        outputs = model(**batch_dict)
        embeddings = average_pool(
            outputs.last_hidden_state, batch_dict["attention_mask"]
        )
        # (Optionally) normalize embeddings
        embeddings = F.normalize(embeddings, p=2, dim=1)
        if tolist:
            embeddings = embeddings.tolist()
        return embeddings

    def __call__(self, **kwargs):
        if "model" not in kwargs:
            raise ValueError("model is required")

        if "mock" in kwargs and kwargs["mock"]:
            import numpy as np

            embeddings = np.array([np.random.rand(768).tolist()])
            return embeddings
        # load files and models, cache it for the next inference
        model_name = kwargs["model"]
        # inference the model
        if model_name == "thenlper/gte-base":
            return self.infer_gte_base_embedding(kwargs["input"])
        else:
            raise ValueError(f"model {model_name} is not supported")


def get_device():
    # Check device availability and set the device
    if torch.cuda.is_available():
        device = torch.device("cuda")
        log.info("Using CUDA (GPU) for inference.")
    elif torch.backends.mps.is_available():
        device = torch.device("mps")
        log.info("Using MPS (Apple Silicon) for inference.")
    else:
        device = torch.device("cpu")
        log.info("Using CPU for inference.")

    return device


def clean_device_cache():
    import torch

    if torch.has_mps:
        torch.mps.empty_cache()

        torch.mps.set_per_process_memory_fraction(1.0)


class TransformerReranker:
    __doc__ = r"""Local model SDK for a reranker model using transformers.

    References:
    - model: https://huggingface.co/BAAI/bge-reranker-base
    - paper: https://arxiv.org/abs/2309.07597

    note:
    If you are using Macbook M1 series chips, you need to ensure ``torch.device("mps")`` is set.
    """
    models: Dict[str, type] = {}

    def __init__(self, model_name: Optional[str] = "BAAI/bge-reranker-base"):
        self.model_name = model_name or "BAAI/bge-reranker-base"
        if model_name is not None:
            self.init_model(model_name=model_name)

    def init_model(self, model_name: str):
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name)
            self.model = AutoModelForSequenceClassification.from_pretrained(model_name)
            # Check device availability and set the device
            device = get_device()

            # Move model to the selected device
            self.device = device
            self.model.to(device)
            self.model.eval()
            # register the model
            self.models[model_name] = self.model  # TODO: better model registration
            log.info(f"Done loading model {model_name}")

        except Exception as e:
            log.error(f"Error loading model {model_name}: {e}")
            raise e

    def infer_bge_reranker_base(
        self,
        # input=List[Tuple[str, str]],  # list of pairs of the query and the candidate
        query: str,
        documents: List[str],
    ) -> List[float]:
        model = self.models.get(self.model_name, None)
        if model is None:
            # initialize the model
            self.init_model(self.model_name)

        # convert the query and documents to pair input
        input = [(query, doc) for doc in documents]

        with torch.no_grad():

            inputs = self.tokenizer(
                input,
                padding=True,
                truncation=True,
                return_tensors="pt",
                max_length=512,
            )
            inputs = {k: v.to(self.device) for k, v in inputs.items()}
            scores = (
                model(**inputs, return_dict=True)
                .logits.view(
                    -1,
                )
                .float()
            )
            # apply sigmoid to get the scores
            scores = F.sigmoid(scores)

        scores = scores.tolist()
        return scores

    def __call__(self, **kwargs):
        r"""Ensure "model" and "input" are in the kwargs."""
        if "model" not in kwargs:
            raise ValueError("model is required")

        # if "mock" in kwargs and kwargs["mock"]:
        #     import numpy as np

        #     scores = np.array([np.random.rand(1).tolist()])
        #     return scores
        # load files and models, cache it for the next inference
        model_name = kwargs["model"]
        # inference the model
        if model_name == self.model_name:
            assert "query" in kwargs, "query is required"
            assert "documents" in kwargs, "documents is required"
            scores = self.infer_bge_reranker_base(kwargs["query"], kwargs["documents"])
            return scores
        else:
            raise ValueError(f"model {model_name} is not supported")


class TransformerLLM:
    __doc__ = r"""Local model SDK for transformers LLM.

    NOTE:
        This inference component is only specific to the HuggingFaceH4/zephyr-7b-beta model.

    The example raw output:
    # <|system|>
    # You are a friendly chatbot who always responds in the style of a pirate.</s>
    # <|user|>
    # How many helicopters can a human eat in one sitting?</s>
    # <|assistant|>
    # Ah, me hearty matey! But yer question be a puzzler! A human cannot eat a helicopter in one sitting, as helicopters are not edible. They be made of metal, plastic, and other materials, not food!


    References:
    - model: https://huggingface.co/HuggingFaceH4/zephyr-7b-beta
    - https://huggingface.co/google/gemma-2b
    - https://huggingface.co/google/gemma-2-2b

    """
    models: Dict[str, type] = {}  # to register the model
    tokenizer: Dict[str, type] = {}

    model_to_init_func = {
        "HuggingFaceH4/zephyr-7b-beta": "use_pipeline",
        "google/gemma-2-2b": "use_pipeline",
    }

    def __init__(
        self,
        model_name: Optional[str] = None,
    ):
        super().__init__()

        self.model_name = model_name  # current model to use

        if model_name is not None and model_name not in self.models:
            self.init_model(model_name=model_name)

    def _check_token(self, token: str):
        import os

        if os.getenv(token) is None:
            warnings.warn(
                f"{token} is not set. You may not be able to access the model."
            )

    def _init_from_pipeline(self, model_name: str):
        from transformers import pipeline

        clean_device_cache()
        self._check_token("HF_TOKEN")
        try:
            import os

            pipe = pipeline(
                "text-generation",
                model=model_name,
                torch_dtype=torch.bfloat16,
                device=get_device(),
                token=os.getenv("HF_TOKEN"),
            )
            self.models[model_name] = pipe
        except Exception as e:
            log.error(f"Error loading model {model_name}: {e}")
            raise e

    def _init_from_automodelcasual_lm(self, model_name: str):
        try:
            from transformers import AutoTokenizer, AutoModelForCausalLM
        except ImportError:
            raise ImportError(
                "transformers is not installed. Please install it with `pip install transformers`"
            )

        try:
            import os

            if os.getenv("HF_TOKEN") is None:
                warnings.warn(
                    "HF_TOKEN is not set. You may not be able to access the model."
                )

            tokenizer = AutoTokenizer.from_pretrained(
                model_name, token=os.getenv("HF_TOKEN")
            )
            model = AutoModelForCausalLM.from_pretrained(
                model_name,
                torch_dtype=torch.bfloat16,
                device_map="auto",
                token=os.getenv("HF_TOKEN"),
            )
            self.models[model_name] = model
            self.tokenizer[model_name] = tokenizer
        except Exception as e:
            log.error(f"Error loading model {model_name}: {e}")
            raise e

    @lru_cache(None)
    def init_model(self, model_name: str):
        log.debug(f"Loading model {model_name}")

        model_setup = self.model_to_init_func.get(model_name, None)
        if model_setup:
            if model_setup == "use_pipeline":
                self._init_from_pipeline(model_name)
            else:
                self._init_from_automodelcasual_lm(model_name)
        else:
            raise ValueError(f"Model {model_name} is not supported")

    def _parse_chat_completion_from_pipeline(self, completion: Any) -> str:

        text = completion[0]["generated_text"]

        pattern = r"(?<=\|assistant\|>).*"

        match = re.search(pattern, text)

        if match:
            text = match.group().strip().lstrip("\\n")
            return text
        else:
            return ""

    def _parse_chat_completion_from_automodelcasual_lm(self, completion: Any) -> str:
        print(f"completion: {completion}")
        return completion[0]

    def parse_chat_completion(self, completion: Any) -> str:
        model_name = self.model_name
        model_setup = self.model_to_init_func.get(model_name, None)
        if model_setup:
            if model_setup == "use_pipeline":
                return self._parse_chat_completion_from_pipeline(completion)
            else:
                return self._parse_chat_completion_from_automodelcasual_lm(completion)
        else:
            raise ValueError(f"Model {model_name} is not supported")

    def _infer_from_pipeline(
        self,
        *,
        model: str,
        messages: Sequence[Dict[str, str]],
        max_tokens: Optional[int] = None,
        **kwargs,
    ):
        if not model:
            raise ValueError("Model is not provided.")

        if model not in self.models:
            self.init_model(model_name=model)

        model_to_use = self.models[model]

        log.info(
            f"Start to infer model {model}, messages: {messages}, kwargs: {kwargs}"
        )

        if model == "HuggingFaceH4/zephyr-7b-beta":

            prompt = model_to_use.tokenizer.apply_chat_template(
                messages, tokenize=False, add_generation_prompt=True
            )

            final_kwargs = {
                "max_new_tokens": max_tokens or 256,
                "do_sample": True,
                "temperature": kwargs.get("temperature", 0.7),
                "top_k": kwargs.get("top_k", 50),
                "top_p": kwargs.get("top_p", 0.95),
            }
            outputs = model_to_use(prompt, **final_kwargs)
        elif model == "google/gemma-2-2b":
            final_kwargs = {
                "max_new_tokens": max_tokens or 256,
                "do_sample": True,
                "temperature": kwargs.get("temperature", 0.7),
                "top_k": kwargs.get("top_k", 50),
                "top_p": kwargs.get("top_p", 0.95),
            }
            text = messages[0]["content"]
            outputs = model_to_use(
                text,
                **final_kwargs,
            )

        log.info(f"Outputs: {outputs}")
        return outputs

    def _infer_from_automodelcasual_lm(
        self,
        *,
        model: str,
        messages: Sequence[Dict[str, str]],
        max_length: Optional[int] = 8192,  # model-agnostic
        **kwargs,
    ):
        if not model:
            raise ValueError("Model is not provided.")
        if model not in self.models:
            self.init_model(model_name=model)
        model_to_use = self.models[model]
        tokenizer_to_use = self.tokenizer[model]

        input_ids = tokenizer_to_use(messages[0]["content"], return_tensors="pt").to(
            get_device()
        )
        print(input_ids)
        outputs_tokens = model_to_use.generate(**input_ids, max_length=max_length)
        outputs = []
        for i, output in enumerate(outputs_tokens):
            outputs.append(tokenizer_to_use.decode(output))
        return outputs

    def infer_llm(
        self,
        *,
        model: str,
        messages: Sequence[Dict[str, str]],
        max_tokens: Optional[int] = None,
        **kwargs,
    ):
        # TODO: generalize the code for more models
        model_setup = self.model_to_init_func.get(model, None)
        if model_setup:
            if model_setup == "use_pipeline":
                return self._infer_from_pipeline(
                    model=model, messages=messages, max_tokens=max_tokens, **kwargs
                )
            else:
                return self._infer_from_automodelcasual_lm(
                    model=model, messages=messages, max_tokens=max_tokens, **kwargs
                )
        else:
            raise ValueError(f"Model {model} is not supported")

    def __call__(self, **kwargs):
        r"""Ensure "model" and "input" are in the kwargs."""
        log.debug(f"kwargs: {kwargs}")
        if "model" not in kwargs:
            raise ValueError("model is required")

        if "messages" not in kwargs:
            raise ValueError("messages is required")

        model_name = kwargs["model"]
        if model_name != self.model_name:
            # need to initialize the model and update the model_name
            self.model_name = model_name
            self.init_model(model_name=model_name)

        output = self.infer_llm(**kwargs)
        return output


class TransformersClient(ModelClient):
    __doc__ = r"""LightRAG API client for transformers.

    Use: ``ls ~/.cache/huggingface/hub `` to see the cached models.

    Some modeles are gated, you will need to their page to get the access token.
    Find how to apply tokens here: https://huggingface.co/docs/hub/security-tokens
    Once you have a token and have access, put the token in the environment variable HF_TOKEN.
    """

    support_models = {
        "thenlper/gte-base": {
            "type": ModelType.EMBEDDER,
        },
        "BAAI/bge-reranker-base": {
            "type": ModelType.RERANKER,
        },
        "HuggingFaceH4/zephyr-7b-beta": {"type": ModelType.LLM},
        "google/gemma-2-2b": {"type": ModelType.LLM},
    }

    def __init__(self, model_name: Optional[str] = None) -> None:
        super().__init__()
        self._model_name = model_name
        if self._model_name:
            assert (
                self._model_name in self.support_models
            ), f"model {self._model_name} is not supported"
        if self._model_name == "thenlper/gte-base":
            self.sync_client = self.init_sync_client()
        elif self._model_name == "BAAI/bge-reranker-base":
            self.reranker_client = self.init_reranker_client()
        elif self._model_name == "HuggingFaceH4/zephyr-7b-beta":
            self.llm_client = self.init_llm_client()
        self.async_client = None

    def init_sync_client(self):
        return TransformerEmbedder()

    def init_reranker_client(self):
        return TransformerReranker()

    def init_llm_client(self):
        return TransformerLLM()

    def set_llm_client(self, llm_client: object):
        r"""Allow user to pass a custom llm client. Here is an example of a custom llm client:

        Ensure you have parse_chat_completion and __call__ methods which will be applied to api_kwargs specified in transform_client.call().

        .. code-block:: python

            class CustomizeLLM:

                def __init__(self) -> None:
                    pass

                def parse_chat_completion(self, completion: Any) -> str:
                    return completion

                def __call__(self, messages: Sequence[Dict[str, str]], model: str, **kwargs):
                    from transformers import AutoTokenizer, AutoModelForCausalLM

                    tokenizer = AutoTokenizer.from_pretrained(
                        "deepseek-ai/deepseek-coder-1.3b-instruct", trust_remote_code=True
                    )
                    model = AutoModelForCausalLM.from_pretrained(
                        "deepseek-ai/deepseek-coder-1.3b-instruct",
                        trust_remote_code=True,
                        torch_dtype=torch.bfloat16,
                    ).to(get_device())
                    messages = [
                        {"role": "user", "content": "write a quick sort algorithm in python."}
                    ]
                    inputs = tokenizer.apply_chat_template(
                        messages, add_generation_prompt=True, return_tensors="pt"
                    ).to(model.device)
                    # tokenizer.eos_token_id is the id of <|EOT|> token
                    outputs = model.generate(
                        inputs,
                        max_new_tokens=512,
                        do_sample=False,
                        top_k=50,
                        top_p=0.95,
                        num_return_sequences=1,
                        eos_token_id=tokenizer.eos_token_id,
                    )
                    print(
                        tokenizer.decode(outputs[0][len(inputs[0]) :], skip_special_tokens=True)
                    )
                    decoded_outputs = []
                    for output in outputs:
                        decoded_outputs.append(
                            tokenizer.decode(output[len(inputs[0]) :], skip_special_tokens=True)
                        )
                    return decoded_outputs

            llm_client = CustomizeLLM()
            transformer_client.set_llm_client(llm_client)
            # use in the generator
            generator = Generator(
                model_client=transformer_client,
                model_kwargs=model_kwargs,
                prompt_kwargs=prompt_kwargs,
                ...)

        """
        self.llm_client = llm_client

    def parse_embedding_response(self, response: Any) -> EmbedderOutput:
        embeddings: List[Embedding] = []
        for idx, emb in enumerate(response):
            embeddings.append(Embedding(index=idx, embedding=emb))
        response = EmbedderOutput(data=embeddings)
        return response

    def parse_chat_completion(self, completion: Any) -> GeneratorOutput:
        try:
            output = self.llm_client.parse_chat_completion(completion)

            return GeneratorOutput(data=output, raw_response=str(completion))
        except Exception as e:
            log.error(f"Error parsing chat completion: {e}")
            return GeneratorOutput(data=None, raw_response=str(completion), error=e)

    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        if "model" not in api_kwargs:
            raise ValueError("model must be specified in api_kwargs")
        if api_kwargs["model"] not in self.support_models:
            raise ValueError(f"model {api_kwargs['model']} is not supported")

        if (
            model_type == ModelType.EMBEDDER
            and "model" in api_kwargs
            and api_kwargs["model"] == "thenlper/gte-base"
        ):
            if self.sync_client is None:
                self.sync_client = self.init_sync_client()
            return self.sync_client(**api_kwargs)
        elif (  # reranker
            model_type == ModelType.RERANKER
            and "model" in api_kwargs
            and api_kwargs["model"] == "BAAI/bge-reranker-base"
        ):
            if not hasattr(self, "reranker_client") or self.reranker_client is None:
                self.reranker_client = self.init_reranker_client()
            scores = self.reranker_client(**api_kwargs)
            top_k_indices, top_k_scores = get_top_k_indices_scores(
                scores, api_kwargs["top_k"]
            )
            return top_k_indices, top_k_scores
        elif model_type == ModelType.LLM and "model" in api_kwargs:  # LLM
            if not hasattr(self, "llm_client") or self.llm_client is None:
                self.llm_client = self.init_llm_client()
            response = self.llm_client(**api_kwargs)
            return response
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    def convert_inputs_to_api_kwargs(
        self,
        input: Any,  # for retriever, it is a single query,
        model_kwargs: dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> dict:
        final_model_kwargs = model_kwargs.copy()
        if model_type == ModelType.EMBEDDER:
            final_model_kwargs["input"] = input
            return final_model_kwargs
        elif model_type == ModelType.RERANKER:
            assert "model" in final_model_kwargs, "model must be specified"
            assert "documents" in final_model_kwargs, "documents must be specified"
            assert "top_k" in final_model_kwargs, "top_k must be specified"
            final_model_kwargs["query"] = input
            return final_model_kwargs
        elif model_type == ModelType.LLM:
            assert "model" in final_model_kwargs, "model must be specified"
            messages = [{"role": "system", "content": input}]
            final_model_kwargs["messages"] = messages
            return final_model_kwargs
        else:
            raise ValueError(f"model_type {model_type} is not supported")


if __name__ == "__main__":
    from adalflow.core import Generator

    import adalflow as adal

    adal.setup_env()

    rag_template = r"""<START_OF_SYSTEM_MESSAGE>
You are a helpful assistant.

Your task is to answer the query that may or may not come with context information.
When context is provided, you should stick to the context and less on your prior knowledge to answer the query.
<END_OF_SYSTEM_MESSAGE>
<START_OF_USER_MESSAGE>
    <START_OF_QUERY>
    {{input_str}}
    <END_OF_QUERY>
    {% if context_str %}
    <START_OF_CONTEXT>
    {{context_str}}
    <END_OF_CONTEXT>
    {% endif %}
<END_OF_USER_MESSAGE>
"""

    template = """{{input_str}}"""

    model_kwargs = {
        "model": "google/gemma-2-2b",
        "temperature": 1,
        "stream": False,
    }
    prompt_kwargs = {
        "input_str": "Where is Brian?",
        # "context_str": "Brian is in the kitchen.",
    }
    prompt_kwargs = {
        "input_str": "What is the capital of France?",
    }

    class CustomizeLLM:

        def __init__(self) -> None:
            pass

        def parse_chat_completion(self, completion: Any) -> str:
            return completion[0]

        def __call__(self, messages: Sequence[Dict[str, str]], model: str, **kwargs):
            r"""take api key"""
            from transformers import AutoTokenizer, AutoModelForCausalLM

            tokenizer = AutoTokenizer.from_pretrained(
                "deepseek-ai/deepseek-coder-1.3b-instruct", trust_remote_code=True
            )
            model = AutoModelForCausalLM.from_pretrained(
                "deepseek-ai/deepseek-coder-1.3b-instruct",
                trust_remote_code=True,
                torch_dtype=torch.bfloat16,
            ).to(get_device())
            messages = [
                {"role": "user", "content": "write a quick sort algorithm in python."}
            ]
            inputs = tokenizer.apply_chat_template(
                messages, add_generation_prompt=True, return_tensors="pt"
            ).to(model.device)
            # tokenizer.eos_token_id is the id of <|EOT|> token
            outputs = model.generate(
                inputs,
                max_new_tokens=512,
                do_sample=False,
                top_k=50,
                top_p=0.95,
                num_return_sequences=1,
                eos_token_id=tokenizer.eos_token_id,
            )

            decoded_outputs = []
            for output in outputs:
                decoded_outputs.append(
                    tokenizer.decode(output[len(inputs[0]) :], skip_special_tokens=True)
                )
            return decoded_outputs

    transformer_client = TransformersClient()
    transformer_client.set_llm_client(CustomizeLLM())
    generator = Generator(
        model_client=transformer_client,
        model_kwargs=model_kwargs,
        # prompt_kwargs=prompt_kwargs,
        template=template,
        # output_processors=JsonParser(),
    )

    output = generator(prompt_kwargs=prompt_kwargs)
    print(output)



"""This is the metric for QA generation. It compares the predicted answer with the ground truth answer."""

from typing import List, Literal
from adalflow.eval.base import BaseEvaluator, EvaluationResult
from adalflow.optim.parameter import Parameter
from adalflow.eval.utils import normalize_answer, f1_score


class AnswerMatchAcc(BaseEvaluator):
    r"""
    Metric for answer matching. It compares the predicted answer with the ground truth answer.

    Args:
        type (str): Type of matching evaluation. Can be "exact_match" or "fuzzy_match". "exact_match" requires the predicted answer to be exactly the same as the ground truth answer. "fuzzy_match" requires the predicted answer to contain the ground truth answer.

    Examples:
        >>> pred_answers = ["positive", "negative", "this is neutral"]
        >>> gt_answers = ["positive", "negative", "neutral"]
        >>> answer_match_acc = AnswerMatchAcc(type="exact_match")
        >>> avg_acc, acc_list = answer_match_acc.compute(all_pred_answer, all_gt_answer)
        >>> avg_acc
        2 / 3
        >>> acc_list
        [1.0, 1.0, 0.0]
        >>> answer_match_acc = AnswerMatchAcc(type="fuzzy_match")
        >>> avg_acc, acc_list = answer_match_acc.compute(all_pred_answer, all_gt_answer)
        >>> avg_acc
        1.0
        >>> acc_list
        [1.0, 1.0, 1.0]

    References:
    1. HotpotQA: https://github.com/hotpotqa/hotpot/blob/master/hotpot_evaluate_v1.py
    """

    def __init__(
        self,
        type: Literal[
            "exact_match",
            "fuzzy_match",
            "rouge_score",
            "bleu_score",
            "bert_score",
            "f1_score",
        ] = "exact_match",
    ):
        self.type = type
        if self.type == "bert_score":
            from torchmetrics.text.bert import BERTScore

            self.bertscore = BERTScore()

        elif self.type == "rouge_score":
            from torchmetrics.text.rouge import ROUGEScore

            self.rougescore = ROUGEScore()

        elif self.type == "bleu_score":
            from torchmetrics.text.bleu import BLEUScore

            self.bleuscore = BLEUScore()

    def compute_single_item(
        self,
        y: object,
        y_gt: object,
    ) -> float:
        r"""
        Compute the match accuracy of the predicted answer for a single query.

        Allow any type of input for pred_answer and gt_answer.
        When evaluating, the input will be converted to string.

        Args:
            pred_answer (object): Predicted answer.
            gt_answer (object): Ground truth answer.

        Returns:
            float: Match accuracy.
        """
        if isinstance(y, Parameter):
            y = y.data
        if isinstance(y_gt, Parameter):
            y_gt = y_gt.data
        try:
            y = str(y).strip()
            y_gt = str(y_gt).strip()
        except Exception as e:
            raise ValueError(
                f"Error converting pred_answer and gt_answer to string: {e}"
            )
        if self.type == "exact_match":
            return 1.0 if normalize_answer(y) == normalize_answer(y_gt) else 0.0
        elif self.type == "fuzzy_match":
            y = normalize_answer(y)
            y_gt = normalize_answer(y_gt)
            return 1.0 if y_gt in y else 0.0
        elif self.type == "f1_score":
            return f1_score(y, y_gt)
        elif self.type == "bert_score":
            from torchmetrics.text.bert import BERTScore

            self.bertscore = BERTScore()
            score = self.bertscore([y], [y_gt])
            # get the data from the tensor
            print(f"y: {[y]}, y_gt: {[y_gt]}, type: {type(y)}, type_gt: {type(y_gt)}")
            print(score)
            single_score = score["precision"].item()
            return single_score
        elif self.type == "rouge_score":
            from torchmetrics.text.rouge import ROUGEScore

            self.rougescore = ROUGEScore()
            score = self.rougescore([y], [y_gt])
            # get the data from the tensor
            print(f"y: {[y]}, y_gt: {[y_gt]}, type: {type(y)}, type_gt: {type(y_gt)}")
            print(score)
            single_score = score["rouge1_precision"].item()
            return single_score
        elif self.type == "bleu_score":
            from torchmetrics.text.bleu import BLEUScore

            self.bleuscore = BLEUScore()
            score = self.bleuscore([y], [y_gt])
            # get the data from the tensor
            print(f"y: {[y]}, y_gt: {[y_gt]}, type: {type(y)}, type_gt: {type(y_gt)}")
            print(score)
            single_score = score.item()
            return single_score

        else:
            raise NotImplementedError

    def compute(
        self, pred_answers: List[str], gt_answers: List[str]
    ) -> EvaluationResult:
        r"""
        Compute the match accuracy of the predicted answer for a list of queries.

        Args:
            pred_answers (List[str]): List of predicted answer strings.
            gt_answers (List[str]): List of ground truth answer strings.

        Returns:
            tuple:
                - float: Average match accuracy.
                - List[float]: Match accuracy values for each query.
        """
        match_acc_list = []
        for pred_answer, gt_answer in zip(pred_answers, gt_answers):
            match = self.compute_single_item(pred_answer, gt_answer)
            match_acc_list.append(match)

        return EvaluationResult(
            avg_score=sum(match_acc_list) / len(match_acc_list),
            per_item_scores=match_acc_list,
        )



from typing import Dict, Any, List, Optional
import os
import logging


from dataclasses import dataclass, field
from datetime import datetime
import json

from adalflow.core.generator import GeneratorOutput
from adalflow.core.base_data_class import DataClass
from adalflow.utils import append_to_jsonl, load_jsonl


log = logging.getLogger(__name__)


@dataclass
class GeneratorCallRecord(DataClass):
    prompt_kwargs: Dict[str, Any] = field(
        default_factory=dict, metadata={"desc": "The prompt kwargs"}
    )
    model_kwargs: Dict[str, Any] = field(
        default_factory=dict, metadata={"desc": "The model kwargs"}
    )
    input: Dict[str, Any] = field(
        default_factory=dict,
        metadata={"desc": "Everything sent to the model api provider"},
    )
    output: GeneratorOutput = field(
        default_factory=GeneratorOutput, metadata={"desc": "The generator output"}
    )
    metadata: Dict[str, Any] = field(
        default_factory=dict, metadata={"desc": "The metadata"}
    )
    time_stamp: str = field(default_factory=str, metadata={"desc": "The time stamp"})


class GeneratorCallLogger:
    __doc__ = r"""Log the generator calls.

    Allow multiple generators to be logged, and each with its own jsonl file.

    The log files are stored in the ./traces/ directory. If a project_name is provided,
    it will be stored in ./traces/{project_name}.

    Args:
        save_dir (str, optional): The directory to save the log files. Defaults to "./traces/".
        project_name (str, optional): The project name. Defaults to None.
    """
    _generator_names_to_files: Dict[str, str] = {}
    # TODO: a project will share the same metadata file to record each logger's metadata
    _metadata_filename = "logger_metadata.json"

    def __init__(
        self, save_dir: Optional[str] = None, project_name: Optional[str] = None
    ):
        self.filepath = save_dir or "./traces/"
        self.project_name = project_name
        if project_name:
            self.filepath = os.path.join(self.filepath, project_name)

        os.makedirs(self.filepath, exist_ok=True)
        self._metadata_filepath = os.path.join(self.filepath, self._metadata_filename)
        if os.path.exists(self._metadata_filepath):
            self.load_meta_data(self._metadata_filepath)
        else:
            self.save_meta_data(self._metadata_filepath)

    def reset(self):
        self._generator_names_to_files = {}

    @property
    def generator_names_to_files(self) -> Dict[str, str]:
        return self._generator_names_to_files

    def get_log_location(self, name: str) -> str:
        return self._generator_names_to_files.get(name, None)

    def get_calls(self, name: str) -> List[GeneratorCallRecord]:
        r"""Get the generator call records for generator with name."""
        return self.load(name)

    def register_generator(self, name: str, filename: Optional[str] = None):
        r"""Register a generator with a name and a jsonl file to log its calls.

        Args:
            name (str): The name of the generator.
            filename (str, optional): The jsonl filename to log the calls. Defaults to {name}_call.jsonl.
        """
        if name in self._generator_names_to_files:
            log.warning(
                f"Generator {name} is already registered with jsonl file at {self._generator_names_to_files[name]}"
            )
            return
        filename = filename or f"{name}_call.jsonl"
        self._generator_names_to_files[name] = os.path.join(self.filepath, filename)
        self.save_meta_data(self._metadata_filepath)

    def save_meta_data(self, file_path: str):
        """Save the _generator_names_to_files to a json file."""
        with open(file_path, "w") as file:
            json.dump(self._generator_names_to_files, file, indent=4)

    def load_meta_data(self, file_path: str):
        """Load the _generator_names_to_files from a json file."""
        if not os.path.exists(file_path):
            log.error(f"File {file_path} does not exist.")
            return

        with open(file_path, "r") as file:
            self._generator_names_to_files = json.load(file)

    def load(self, name: str) -> List[GeneratorCallRecord]:
        r"""Load the generator call records."""
        if name not in self._generator_names_to_files:
            log.error(f"Generator {name} is not registered.")
            raise FileNotFoundError(f"Generator {name} is not registered.")

        records: List[Dict] = load_jsonl(self._generator_names_to_files[name])
        return [GeneratorCallRecord.from_dict(record) for record in records]

    def log_call(
        self,
        name: str,
        output: GeneratorOutput,
        input: Optional[Dict[str, Any]] = None,
        prompt_kwargs: Optional[Dict[str, Any]] = None,
        model_kwargs: Optional[Dict[str, Any]] = None,
    ):
        r"""Log the generator call."""

        if name not in self._generator_names_to_files:
            log.error(f"Generator {name} is not registered.")
            return

        record = GeneratorCallRecord(
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
            input=input,
            output=output,
            time_stamp=datetime.now().isoformat(),
        )
        append_to_jsonl(self._generator_names_to_files[name], record.to_dict())



"""Parameter is used by Optimizer, Trainers, AdalComponent to auto-optimizations"""

from typing import (
    Generic,
    TypeVar,
    Any,
    List,
    Set,
    Dict,
    Tuple,
    Optional,
    Literal,
    Callable,
    TYPE_CHECKING,
)
from collections import defaultdict
import logging
import os
from dataclasses import dataclass, field
import uuid
from adalflow.optim.types import ParameterType
from adalflow.core.base_data_class import DataClass
from adalflow.utils.logger import printc
import html
from adalflow.optim.gradient import Gradient


if TYPE_CHECKING:
    from adalflow.optim.text_grad.tgd_optimizer import TGDData, TGDOptimizerTrace

T = TypeVar("T")  # covariant set to False to allow for in-place updates

log = logging.getLogger(__name__)

__all__ = ["Parameter", "ComponentNode", "ComponentTrace", "ScoreTrace"]


@dataclass
class ComponentTrace(DataClass):
    name: str = field(metadata={"desc": "The name of the component"}, default=None)
    id: str = field(metadata={"desc": "The unique id of the component"}, default=None)
    input_args: Dict[str, Any] = field(
        metadata={"desc": "The input arguments of the GradComponent forward"},
        default=None,
    )
    full_response: object = field(
        metadata={"desc": "The full response of the GradComponent output"}, default=None
    )
    raw_response: str = field(
        metadata={"desc": "The raw response of the generator"}, default=None
    )
    api_kwargs: Dict[str, Any] = field(
        metadata={
            "desc": "The api_kwargs for components like Generator and Retriever that pass to the model client"
        },
        default=None,
    )

    def to_context_str(self):
        output = f"""<INPUT>: {self.input_args}. <OUTPUT>: {self.full_response}"""
        return output


# TODO: use this to better trace the score
@dataclass
class ScoreTrace:
    score: float = field(metadata={"desc": "The score of the data point"}, default=None)
    eval_comp_id: str = field(
        metadata={"desc": "The id of the evaluation component"}, default=None
    )
    eval_comp_name: str = field(
        metadata={"desc": "The name of the evaluation component"}, default=None
    )


@dataclass(frozen=True)
class ComponentNode(DataClass):
    """Used to represent a node in the component graph."""

    id: str = field(metadata={"desc": "The unique id of the component"})
    name: str = field(metadata={"desc": "The name of the component"})
    type: Literal["INPUT", "COMPONENT"] = field(
        metadata={"desc": "The type of the node"}, default="COMPONENT"
    )


COMBINED_GRADIENTS_TEMPLATE = r"""
{% if component_schema %}
<COMPONENT_SCHEMA>
Gradients are from {{ component_schema | length }} components.
{% for component_id, schema in component_schema.items() %}
{{ schema }}
{% endfor %}
</COMPONENT_SCHEMA>
{% endif %}

{% if combined_gradients %}
{% for group in combined_gradients %}
<DataID: {{ loop.index }} >
<AVERAGE_SCORE>{{ group.average_score|round(2) }}</AVERAGE_SCORE>
{% for gradient in group.gradients %}
{{ loop.index }}.
INPUT_OUTPUT: {{ gradient.context }}
{% if gradient.score is not none %}
<SCORE>{{ gradient.score | round(3) }}</SCORE>
{% endif %}
{% if gradient.gradient is not none %}
<FEEDBACK>{{ gradient.gradient }}</FEEDBACK>
{% endif %}
{% endfor %}
</DataID>
{% endfor %}
{% endif %}
"""

# id: {{ component_id }}, remove using component id


class Parameter(Generic[T]):
    r"""A data container to represent a parameter used for optimization.

    A parameter enforce a specific data type and can be updated in-place.
    When parameters are used in a component - when they are assigned as Component attributes
    they are automatically added to the list of its parameters, and  will
    appear in the :meth:`~Component.parameters` or :meth:`~Component.named_parameters` method.

    Args:

    End users only need to create the Parameter with four arguments and pass it to the
    prompt_kwargs in the Generator.
        - data (str): the data of the parameter
        - requires_opt (bool, optional): if the parameter requires optimization. Default: `True`
        - role_desc.
        - param_type, incuding ParameterType.PROMPT for instruction optimization, ParameterType.DEMOS
        for few-shot optimization.
        - instruction_to_optimizer (str, optional): instruction to the optimizer. Default: `None`
        - instruction_to_backward_engine (str, optional): instruction to the backward engine. Default: `None`

    The parameter users created will be automatically assigned to the variable_name/key in the prompt_kwargs
    for easy reading and debugging in the trace_graph.

    References:

    1. https://github.com/karpathy/micrograd/blob/master/micrograd/engine.py
    """

    allowed_types = {
        ParameterType.NONE,
        ParameterType.PROMPT,
        ParameterType.DEMOS,
        ParameterType.HYPERPARAM,
        ParameterType.INPUT,
    }

    id: str = None  # Unique id of the parameter
    name: str = None  # Name of the parameter, easier to read for humans
    role_desc: str = ""  # Description of the role of the parameter
    data: T = None  # Data of the parameter
    data_id: str = (
        None  # Id of the data from the training set, used only for input_type
    )
    param_type: ParameterType

    proposing: bool = False  # State of the parameter
    predecessors: Set["Parameter"] = set()  # Predecessors of the parameter
    peers: Set["Parameter"] = set()  # Peers of the parameter
    eval_input: object = None  # Eval input passing to the eval_fn or evaluator you use
    successor_map_fn: Dict[str, Callable] = (
        None  # Map function to get the data from the output
    )

    tgd_optimizer_trace: "TGDOptimizerTrace" = None  # Trace of the TGD optimizer

    data_in_prompt: Callable = (
        None  # Callable to get the str of the data to be used in the prompt
    )
    gt: object = None  # Ground truth of the parameter

    def __init__(
        self,
        *,
        id: Optional[str] = None,  # unique id of the parameter
        data: T = None,  # for generator output, the data will be set up as raw_response
        data_id: str = None,  # for tracing the data item in the training/val/test set
        requires_opt: bool = True,
        role_desc: str = "",
        param_type: ParameterType = ParameterType.NONE,
        name: str = None,  # name is used to refer to the parameter in the prompt, easier to read for humans
        instruction_to_optimizer: str = None,
        instruction_to_backward_engine: str = None,
        score: Optional[float] = None,
        eval_input: object = None,
        successor_map_fn: Optional[Dict[str, Callable]] = None,
        data_in_prompt: Callable = None,
    ):
        self.id = id or str(uuid.uuid4())
        self.data_id = data_id

        self.name = name
        self.role_desc = role_desc

        if not self.name:
            self.name = (
                self.role_desc.capitalize().replace(" ", "_")[0:10]
                if self.role_desc
                else f"param_{self.id}"
            )
        self.param_type = param_type
        # allow subclasses to override allowed_types dynamically
        allowed_types = getattr(self.__class__, "allowed_types", set())
        if param_type not in allowed_types:
            raise ValueError(
                f"{param_type.name} is not allowed for {self.__class__.__name__}"
            )

        self.data = data  # often string and will be used in the prompts
        self.requires_opt = requires_opt
        self.data_type = type(data)

        self.set_eval_fn_input(eval_input=data)
        self.gradients: Set[Gradient] = set()

        self.grad_fn = None

        self.previous_data = None  # used to store the previous data
        # context of the forward pass

        self.instruction_to_optimizer: str = instruction_to_optimizer
        self.instruction_to_backward_engine: str = instruction_to_backward_engine

        # here are used for demo parameter, filled by generator.forward
        self._traces: Dict[str, DataClass] = {}  # id to data items (DynamicDataClass)
        self._student_traces: Dict[str, DataClass] = {}  # id

        self.score: float = (
            score  # end to end evaluation score, TODO: might have multiple scores if using multiple eval fns  # score is set in the gradients in the backward pass
        )

        self._demos: List[DataClass] = (
            []
        )  # used for the optimizer to save the proposed demos
        self._previous_demos: List[DataClass] = []
        self.eval_input = eval_input

        self.successor_map_fn = successor_map_fn or {}

        def default_prompt_map_fn(param: Parameter):
            # if isinstance(param.data, GeneratorOutput):
            #     return param.data.raw_response
            return param.data

        self.data_in_prompt = data_in_prompt or default_prompt_map_fn
        self.gt = None

    def map_to_successor(self, successor: object) -> T:
        """Apply the map function to the successor based on the successor's id."""
        successor_id = id(successor)
        if successor_id not in self.successor_map_fn:
            default_map_fn = lambda x: x.data  # noqa: E731
            return default_map_fn(self)

        return self.successor_map_fn[successor_id](self)

    def add_successor_map_fn(self, successor: object, map_fn: Callable):
        """Add or update a map function of the value for a specific successor using its id.
        succssor will know the value of the current parameter."""
        self.successor_map_fn[id(successor)] = map_fn

    def check_if_already_computed_gradient_respect_to(self, response_id: str) -> bool:
        from_response_ids = [g.from_response_id for g in self.gradients]
        return response_id in from_response_ids

    ############################################################################################################
    # Handle gt
    ############################################################################################################
    def set_gt(self, gt: object):

        self.gt = gt

    def get_gt(self) -> object:
        return self.gt

    # ############################################################################################################
    # Handle gradients and context
    # ############################################################################################################
    def add_gradient(self, gradient: "Gradient"):
        # if gradient.param_type != ParameterType.GRADIENT:
        #     raise ValueError("Cannot add non-gradient parameter to gradients list.")

        if gradient.from_response_id is None:
            raise ValueError("Gradient must have a from_response_id.")

        start_order = len(self.gradients)
        gradient.order = start_order

        self.gradients.add(gradient)
        # sort the gradients by the data_id, response_component_id, and score
        self.sort_gradients()

    def reset_gradients(self):
        self.gradients = set()

    def get_gradients_names(self) -> str:
        names = [g.name for g in self.gradients]
        names = ", ".join(names)
        return names

    def get_prompt_data(self) -> str:
        return self.data_in_prompt(self)

    def get_gradients_str(self) -> str:
        if not self.gradients:
            return ""

        gradients_str = ""
        for i, g in enumerate(self.gradients):
            gradients_str += f"{i}. {g.data}\n"

        return gradients_str

    def get_gradient_and_context_text(self, skip_correct_sample: bool = False) -> str:
        """Aggregates and returns:
        1. the gradients
        2. the context text for which the gradients are computed

        Sort the gradients from the lowest score to the highest score.
        Highlight the gradients with the lowest score to the optimizer.
        """
        from adalflow.core.prompt_builder import Prompt

        if not self.gradients:
            return ""

        # print the score for the sorted gradients
        lowest_score_gradients = []
        for i, g in enumerate(self.gradients):
            if skip_correct_sample:
                if g.score > 0.5:
                    continue
            lowest_score_gradients.append(g)

        gradient_context_combined_str = ""
        if lowest_score_gradients and len(lowest_score_gradients) > 0:

            # group gradients by data_id and calculate average scores
            grouped_gradients = defaultdict(
                lambda: {"gradients": [], "score_sum": 0, "count": 0}
            )
            for g in lowest_score_gradients:
                group = grouped_gradients[g.data_id]
                group["gradients"].append(
                    {
                        "gradient": g.data,
                        "context": g.context.input_output,
                        "score": g.score,
                    }
                )
                group["score_sum"] += g.score if g.score is not None else 0
                group["count"] += 1

            # Calculate average scores and sort groups
            grouped_list = []
            for data_id, group in grouped_gradients.items():
                average_score = (
                    group["score_sum"] / group["count"] if group["count"] > 0 else 0
                )
                grouped_list.append(
                    {
                        "data_id": data_id,
                        "average_score": average_score,
                        "gradients": group["gradients"],
                    }
                )
            sorted_groups = sorted(grouped_list, key=lambda x: x["average_score"])

            gradient_context_combined_str = Prompt(
                template=COMBINED_GRADIENTS_TEMPLATE,
                prompt_kwargs={"combined_gradients": sorted_groups},
            )().strip()

        # get component id: gradient
        component_id_to_gradient: Dict[str, Gradient] = {}
        for g in lowest_score_gradients:
            component_id_to_gradient[g.from_response_component_id] = g

        componend_id_to_schema: Dict[str, str] = {}
        for id, g in component_id_to_gradient.items():
            componend_id_to_schema[id] = g.context.to_yaml(exclude={"input_output"})

        # if there are multiple successors, there will be multiple component schemas

        return gradient_context_combined_str

    def get_gradients_component_schema(self, skip_correct_sample: bool = False) -> str:
        """Aggregates and returns:
        1. the gradients
        2. the context text for which the gradients are computed

        Sort the gradients from the lowest score to the highest score.
        Highlight the gradients with the lowest score to the optimizer.
        """
        from adalflow.core.prompt_builder import Prompt

        # print(
        #     f"len of gradients: {len(self.gradients)}, scores: {[g._score for g in self.gradients]} for {self.name}"
        # )

        if not self.gradients:
            return ""

        # sore gradients by the _score from low to high
        # self.gradients = sorted(
        #     self.gradients, key=lambda x: x.score if x.score is not None else 1
        # )
        # print the score for the sorted gradients
        lowest_score_gradients = []
        for i, g in enumerate(self.gradients):
            if skip_correct_sample:
                if g.score > 0.5:
                    continue
            lowest_score_gradients.append(g)

        # Group gradients by `data_id` and calculate average scores
        grouped_gradients = defaultdict(
            lambda: {"gradients": [], "score_sum": 0, "count": 0}
        )
        for g in lowest_score_gradients:
            group = grouped_gradients[g.data_id]
            group["gradients"].append(
                {
                    "gradient": g.data,
                    "context": g.context.input_output,
                    "score": g.score,
                }
            )
            group["score_sum"] += g.score if g.score is not None else 0
            group["count"] += 1

        # Calculate average scores and sort groups
        grouped_list = []
        for data_id, group in grouped_gradients.items():
            average_score = (
                group["score_sum"] / group["count"] if group["count"] > 0 else 0
            )
            grouped_list.append(
                {
                    "data_id": data_id,
                    "average_score": average_score,
                    "gradients": group["gradients"],
                }
            )
        sorted_groups = sorted(grouped_list, key=lambda x: x["average_score"])

        # get component id: gradient
        component_id_to_gradient: Dict[str, Gradient] = {}
        for g in lowest_score_gradients:
            component_id_to_gradient[g.from_response_component_id] = g

        componend_id_to_schema: Dict[str, str] = {}
        for id, g in component_id_to_gradient.items():
            componend_id_to_schema[id] = g.context.to_yaml(exclude=["input_output"])

        # parse the gradients and context.
        gradients_and_context: List[Dict[str, Any]] = (
            []
        )  # {gradient: data, context: GradientContext.input_output}
        for g in lowest_score_gradients:
            gradients_and_context.append(
                {
                    "data_id": g.data_id,
                    "gradient": g.data,
                    "context": g.context.input_output,
                    "score": g.score,
                }
            )

        gradient_context_combined_str = Prompt(
            template=COMBINED_GRADIENTS_TEMPLATE,
            prompt_kwargs={
                "combined_gradients": sorted_groups,
                "component_schema": componend_id_to_schema,
            },
        )().strip()

        # if there are multiple successors, there will be multiple component schemas

        return gradient_context_combined_str

    def merge_gradients_for_cycle_components(self):
        """Merge data_id, from_response_component_id into the same gradient"""

    def sort_gradients(self):
        """With rules mentioned in Graient class, we will track the gradients by data_id, then response_component_id, then score"""

        self.gradients = sorted(
            self.gradients,
            key=lambda x: (
                x.data_id,
                x.from_response_component_id,
                -x.order if x.order is not None else 0,
                x.from_response_id,
                x.score,
            ),
        )
        # make it a set again
        self.gradients = set(self.gradients)

    ############################################################################################################
    # Setters and getters
    ############################################################################################################

    def set_predecessors(self, predecessors: List["Parameter"] = None):
        if predecessors is None:
            self.predecessors = set()
        else:
            for pred in self.predecessors:
                if not isinstance(pred, Parameter):
                    raise TypeError(
                        f"Expected a list of Parameter instances, got {type(pred).__name__}, {pred}"
                    )
            self.predecessors = set(predecessors)

    def set_grad_fn(self, grad_fn):
        self.grad_fn = grad_fn

    def get_param_info(self):
        """Used to represent the parameter in the prompt."""
        return {
            "name": self.name,
            "role_desc": self.role_desc,
            "prompt_data": self.get_prompt_data(),  # default to use all data
            "param_type": self.param_type,
            "requires_opt": self.requires_opt,
            "eval_input": self.eval_input,  # for output passing to the eval_fn
        }

    def set_peers(self, peers: List["Parameter"] = None):
        if peers is None:
            self.peers = set()
        else:
            for peer in peers:
                if not isinstance(peer, Parameter):
                    raise TypeError(
                        f"Expected a list of Parameter instances, got {type(peer).__name__}, {peer}"
                    )
            self.peers = set(peers)

    #############################################################################################################
    # Trace the tgd optimizer data
    ############################################################################################################
    def trace_optimizer(self, api_kwargs: Dict[str, Any], response: "TGDData"):
        r"""Trace the inputs and output of a TGD optimizer."""
        from adalflow.optim.text_grad.tgd_optimizer import TGDOptimizerTrace

        self.tgd_optimizer_trace = TGDOptimizerTrace(
            api_kwargs=api_kwargs, output=response
        )

    def set_eval_fn_input(self, eval_input: object):
        r"""Set the input for the eval_fn."""
        self.eval_input = eval_input

    ###################################################################################################################
    #   Used for demo optimizer (forward and backward pass) to accumlate the traces on both score and DynamicDataClass
    ###################################################################################################################
    def set_score(self, score: float):
        r"""Set the score of the parameter in the backward pass
        For intermediate nodes, there is only one score per each eval fn behind this node.
        For leaf nodes, like DEMO or PROMPT, it will have [batch_size] of scores.

        But this score is only used to relay the score to the demo parametr.
        """
        score = float(score)
        if not isinstance(score, float):
            raise ValueError(
                f"score is not float, but {type(score)}, parameter name: {self.name}"
            )
        self.score = score

    def add_dataclass_to_trace(self, trace: DataClass, is_teacher: bool = True):
        r"""Called by the generator.forward to add a trace to the parameter.

        It is important to allow updating to the trace, as this will give different sampling weight.
        If the score increases as the training going on, it will become less likely to be sampled,
        allowing the samples to be more diverse. Or else, it will keep sampling failed examples.
        """
        target = self._traces if is_teacher else self._student_traces
        if not hasattr(trace, "id"):
            raise ValueError("Trace must have an id attribute.")
        if trace.id in target:
            print(f"Trace with id {trace.id} already exists. Updating the trace.")
        target[trace.id] = trace

    def add_score_to_trace(self, trace_id: str, score: float, is_teacher: bool = True):
        r"""Called by the generator.backward to add the eval score to the trace."""

        target = self._traces if is_teacher else self._student_traces

        if trace_id not in target:
            raise ValueError(
                f"Trace with id {trace_id} does not exist. Current traces: {target.keys()}"
            )

        setattr(target[trace_id], "score", score)

        from adalflow.utils.logger import printc

        printc(f"Adding score {score} to trace {trace_id}", "magenta")

    ############################################################################################################
    #   Used for optimizer to propose new data
    ############################################################################################################
    def propose_data(self, data: T, demos: Optional[List[DataClass]] = None):
        r"""Used by optimizer to put the new data, and save the previous data in case of revert."""
        if self.proposing:
            raise ValueError("Cannot propose a new data when it is already proposing.")
        self.previous_data = self.data
        self.data = data
        self.proposing = True
        if demos is not None:
            self._previous_demos = self._demos
            self._demos = demos

    def revert_data(self, include_demos: bool = False):
        r"""Revert the data to the previous data."""
        if not self.proposing:
            raise ValueError("Cannot revert data without proposing first.")

        self.data = self.previous_data
        self.previous_data = None
        self.proposing = False

        # cant reset gradients yet for the loss
        if include_demos:
            self._demos = self._previous_demos
            self._previous_demos = []

    def step_data(self, include_demos: bool = False):
        r"""Use PyTorch's optimizer syntax to finalize the update of the data."""
        if not self.proposing:
            raise ValueError("Cannot set data without proposing first.")

        self.previous_data = None
        self.proposing = False

        if include_demos:
            self._previous_demos = []

    def get_grad_fn(self):
        return self.grad_fn

    def _check_data_type(self, new_data: Any):
        """Check the type of new_data against the expected data type."""
        if self.data is not None and not isinstance(new_data, self.data_type):
            raise TypeError(
                f"Expected data type {self.data_type.__name__}, got {type(new_data).__name__}"
            )

    def update_value(self, data: T):
        """Update the parameter's value in-place, checking for type correctness."""
        self._check_data_type(data)
        if self.data is None and data is not None:
            self.data_type = type(data)
        self.data = data

    # TODO: dont use short value
    def get_short_value(self, n_words_offset: int = 10) -> str:
        """
        Returns a short version of the value of the variable. We sometimes use it during optimization, when we want to see the value of the variable, but don't want to see the entire value.
        This is sometimes to save tokens, sometimes to reduce repeating very long variables, such as code or solutions to hard problems.
        :param n_words_offset: The number of words to show from the beginning and the end of the value.
        :type n_words_offset: int
        """
        # 1. ensure the data is a string
        # data = self.data
        data = self.get_prompt_data()
        if not isinstance(self.data, str):
            data = str(self.data)
        words = data.split(" ")
        if len(words) <= 2 * n_words_offset:
            return data
        short_value = (
            " ".join(words[:n_words_offset])
            + " (...) "
            + " ".join(words[-n_words_offset:])
        )
        return short_value

    def reset_all_gradients(self):
        """Traverse the graph and reset the gradients for all nodes."""
        nodes, _ = Parameter.trace_graph(self)
        for node in nodes:
            node.reset_gradients()

    @staticmethod
    def trace_graph(
        root: "Parameter",
    ) -> Tuple[Set["Parameter"], Set[Tuple["Parameter", "Parameter"]]]:
        nodes, edges = set(), set()

        def build_graph(node: "Parameter"):
            if node in nodes:
                return
            if node is None:
                raise ValueError("Node is None")
            nodes.add(node)
            for pred in node.predecessors:
                edges.add((pred, node))
                build_graph(pred)

        build_graph(root)
        return nodes, edges

    def backward(
        self,
    ):
        """
        Apply backward pass for for all nodes in the graph by reversing the topological order.
        """
        # engine should be the llm or customized backwards function to pass feedback

        # topological sort of all the predecessors of the current parameter in the graph
        log.debug(f"Backward pass for {self.data}, backward function: {self.grad_fn}")
        topo: List[Parameter] = []
        visited = set()

        def build_topo(node: Parameter):
            if node in visited:
                return
            visited.add(node)
            for pred in node.predecessors:
                build_topo(pred)
            topo.append(node)

        build_topo(self)
        # backpropagation

        self.gradients = set()
        for node in reversed(topo):
            if not node.requires_opt:
                log.debug(f"Skipping {node.name} as it does not require optimization")
                continue
            component_name = None
            if hasattr(node, "component_trace"):
                component_name = node.component_trace.name
            printc(
                f"node: {node.name}, component: {component_name}, grad_fn: {node.grad_fn}."
            )
            if node.get_grad_fn() is not None:  # gradient function takes in the engine
                log.debug(f"Calling gradient function for {node.name}")
                node.grad_fn()

    @staticmethod
    def generate_node_html(node: "Parameter", output_dir="node_pages"):
        """Generate an HTML page for a specific node."""
        import json

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        filename = f"{output_dir}/{node.name}.html"

        # Gather gradients as JSON objects
        gradients = []
        for i, g in enumerate(node.gradients):
            gradient = g.to_json_obj()
            for k, v in gradient.items():
                if isinstance(v, str):
                    gradient[k] = v.replace("<", "&lt;").replace(">", "&gt;")
            gradients.append(gradient)

        data_json = None
        node_data_type = str(type(node.data)).replace("<", "&lt;").replace(">", "&gt;")
        printc(f"Node data type: {node_data_type}")
        if isinstance(node.data, dict):
            data_json = data_json
        elif isinstance(node.data, DataClass):
            try:
                data_json = node.data.to_json_obj()
            except Exception:

                data_json = str(node.data)

        else:
            data_json = str(node.data)
            data_json = {"data": data_json}

        gradients_json = json.dumps(gradients, indent=4, ensure_ascii=False)

        optimizer_trace = None
        if node.tgd_optimizer_trace:
            optimizer_trace = node.tgd_optimizer_trace.to_json_obj()
            optimizer_trace = json.dumps(optimizer_trace, indent=4, ensure_ascii=False)

        with open(filename, "w") as file:
            file.write(
                f"""
            <!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <meta name="viewport" content="width=device-width, initial-scale=1.0">
                <title>{node.name}</title>
                <style>
                    pre {{
                        background-color: #f5f5f5;
                        padding: 10px;
                        border-radius: 5px;
                        overflow-x: auto;
                    }}
                </style>
            </head>
            <body>
                <h1>Details for Node: {node.name}</h1>
                <p><b>ID:</b> {node.id}</p>
                <p><b>Role:</b> {node.role_desc}</p>
                <p><b>DataType:</b> {node_data_type}</p>
                <pre><b>Data:</b> \n{json.dumps(data_json, indent=4)}</pre>
                <p><b>Data ID:</b> {node.data_id}</p>
                <p><b>Previous Value:</b> {node.previous_data}</p>
                <p><b>Requires Optimization:</b> {node.requires_opt}</p>
                <p><b>Type:</b> {node.param_type.value} ({node.param_type.description})</p>
                <pre><b>Gradients:</b>\n{gradients_json}</pre>
                <pre><b>TGD Optimizer Trace:</b>\n{optimizer_trace}</pre>

            </body>
            </html>
            """
            )
        print(f"Generated HTML for node: {node.name} at {filename}")

    def draw_interactive_html_graph(
        self,
        filepath: Optional[str] = None,
        nodes: List["Parameter"] = None,
        edges: List[Tuple["Parameter", "Parameter"]] = None,
    ) -> Dict[str, Any]:
        """
        Generate an interactive graph with pyvis and save as an HTML file.

        Args:
            nodes (list): A list of Parameter objects.
            edges (list): A list of edges as tuples (source, target).
            filepath (str, optional): Path to save the graph file. Defaults to None.

        Returns:
            dict: A dictionary containing the graph file path.
        """
        from jinja2 import Template
        from pyvis.network import Network

        output_file = "interactive_graph.html"
        filepath = filepath or "output"
        os.makedirs(filepath, exist_ok=True)
        final_file = os.path.join(filepath, output_file)

        net = Network(height="750px", width="100%", directed=True)

        node_colors = {
            ParameterType.PROMPT: "lightblue",
            ParameterType.DEMOS: "orange",
            ParameterType.INPUT: "gray",
            ParameterType.OUTPUT: "green",
            ParameterType.GENERATOR_OUTPUT: "purple",
            ParameterType.RETRIEVER_OUTPUT: "red",
            ParameterType.LOSS_OUTPUT: "pink",
            ParameterType.SUM_OUTPUT: "blue",
        }

        # Add nodes to the graph
        node_ids = set()
        for node in nodes:
            self.generate_node_html(node, output_dir=filepath)

            node_id = node.id
            node_show_name = node.name.replace(f"_{node_id}", "")
            label = (
                f"""<div style="max-height: 150px; overflow-y: auto; border: 1px solid #ccc; padding: 10px; background: white; position: relative; font-family: Arial, sans-serif;">"""
                f"<b>Name:</b> {node_show_name}<br>"
                f"<b>Role:</b> {node.role_desc.capitalize()}<br>"
                f"<b>Value:</b> {node.data}<br>"
                f"<b>Data ID:</b> {node.data_id}<br>"
            )
            if node.proposing:
                label += "<b>Proposing:</b> Yes<br>"
                label += f"<b>Previous Value:</b> {node.previous_data}<br>"
            label += f"<b>Requires Optimization:</b> {node.requires_opt}<br>"
            if node.param_type:
                label += f"<b>Type:</b> {node.param_type.value}<br>"

            net.add_node(
                n_id=node.id,
                label=node_show_name,
                title=label,
                color=node_colors.get(node.param_type, "gray"),
                url=f"./{node.name}.html",  # Relative path
            )
            node_ids.add(node.id)

        # Add edges to the graph
        for source, target in edges:
            if source.id in node_ids and target.id in node_ids:
                net.add_edge(source.id, target.id)
            else:
                print(
                    f"Skipping edge from {source.name} to {target.name} as one of the nodes does not exist."
                )

        net.toggle_physics(True)
        net.template = Template(
            """
    <!DOCTYPE html>
    <html>
    <head>
        <script src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.min.js"></script>
        <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis-network.min.css" rel="stylesheet" />
        <style>
            #tooltip {
                display: none;
                position: absolute;
                max-width: 300px;
                border: 1px solid #ccc;
                padding: 10px;
                background: white;
                z-index: 1000;
                font-family: Arial, sans-serif;
                font-size: 12px;
                line-height: 1.5;
            }
            #tooltip button {
                display: block;
                margin-top: 10px;
            }
             /* Simple styling for the legend */
            #legend {
                margin-top: 20px;
                font-family: Arial, sans-serif;
                font-size: 14px;
            }
            .legend-item {
                margin-bottom: 5px;
            }
            .legend-color-box {
                display: inline-block;
                width: 12px;
                height: 12px;
                margin-right: 5px;
                border: 1px solid #000;
                vertical-align: middle;
            }
        </style>
    </head>
    <body>
        <div id="tooltip">
            <div id="tooltip-content"></div>
            <button onclick="document.getElementById('tooltip').style.display='none'">Close</button>
        </div>
         <!-- Legend Section -->
        <div id="legend">
            <strong>Legend:</strong>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: lightblue;"></span>PROMPT
            </div>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: orange;"></span>DEMOS
            </div>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: gray;"></span>INPUT
            </div>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: green;"></span>OUTPUT
            </div>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: purple;"></span>GENERATOR_OUTPUT
            </div>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: red;"></span>RETRIEVER_OUTPUT
            </div>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: pink;"></span>LOSS_OUTPUT
            </div>
            <div class="legend-item">
                <span class="legend-color-box" style="background-color: blue;"></span>SUM_OUTPUT
            </div>
        </div>
        <!-- End Legend Section -->
        <div id="mynetwork" style="height: {{ height }};"></div>
        <script type="text/javascript">
            var nodes = new vis.DataSet({{ nodes | safe }});
            var edges = new vis.DataSet({{ edges | safe }});
            var container = document.getElementById('mynetwork');
            var data = { nodes: nodes, edges: edges };
            var options = {{ options | safe }};
            var network = new vis.Network(container, data, options);

            // Handle node click to open a link
            network.on("click", function (params) {
                if (params.nodes.length > 0) {
                    const nodeId = params.nodes[0];
                    const node = nodes.get(nodeId);
                    if (node.url) {
                        window.open(node.url, '_blank');
                    }
                }
            });
        </script>
    </body>
    </html>
    """
        )

        net.show(final_file)
        print(f"Interactive graph saved to {final_file}")

        return {"graph_path": final_file}

    @staticmethod
    def wrap_and_escape(text, width=40):
        r"""Wrap text to the specified width, considering HTML breaks, and escape special characters."""
        try:
            import textwrap
        except ImportError as e:
            raise ImportError(
                "Please install textwrap using 'pip install textwrap' to use this feature"
            ) from e

        def wrap_text(text, width):
            """Wrap text to the specified width, considering HTML breaks."""
            lines = textwrap.wrap(
                text, width, break_long_words=False, replace_whitespace=False
            )
            return "<br/>".join(lines)

        if not isinstance(text, str):
            text = str(text)
        text = (
            text.replace("&", "&amp;")
            .replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace('"', "&quot;")
            .replace("'", "&apos;")
            .replace(
                "\n", "<br/>"
            )  # Convert newlines to HTML line breaks if using HTML labels
        )
        return wrap_text(text, width)

    def draw_graph(
        self,
        add_grads: bool = True,
        full_trace: bool = False,
        format: Literal["png", "svg"] = "png",
        rankdir: Literal["LR", "TB"] = "TB",
        filepath: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Draw the graph of the parameter and its gradients.

        Args:
            add_grads (bool, optional): Whether to add gradients to the graph. Defaults to True.
            format (str, optional): The format of the output file. Defaults to "png".
            rankdir (str, optional): The direction of the graph. Defaults to "TB".
            filepath (str, optional): The path to save the graph. Defaults to None.
            full_trace (bool, optional): Whether to include more detailed trace such as api_kwargs. Defaults to False.
        """
        from adalflow.utils import save_json
        from adalflow.utils.global_config import get_adalflow_default_root_path

        try:
            from graphviz import Digraph

        except ImportError as e:
            raise ImportError(
                "Please install graphviz using 'pip install graphviz' to use this feature"
            ) from e

        assert rankdir in ["LR", "TB"]

        root_path = get_adalflow_default_root_path()

        filename = f"trace_graph_{self.name}_id_{self.id}"
        filepath = (
            os.path.join(filepath, filename)
            if filepath
            else os.path.join(root_path, "graphs", filename)
        )
        # final_path = f"{filepath}.{format}"
        print(f"Saving graph to {filepath}.{format}")

        nodes, edges = self.trace_graph(self)
        dot = Digraph(format=format, graph_attr={"rankdir": rankdir, "dpi": "300"})
        node_names = set()
        for n in nodes:
            label_color = "darkblue"

            node_label = (
                f"<table border='0' cellborder='1' cellspacing='0'>"
                f"<tr><td><b><font color='{label_color}'>Name: </font></b></td><td>{self.wrap_and_escape(n.id)}</td></tr>"
                f"<tr><td><b><font color='{label_color}'>Name: </font></b></td><td>{self.wrap_and_escape(n.name)}</td></tr>"
                f"<tr><td><b><font color='{label_color}'>Role: </font></b></td><td>{self.wrap_and_escape(n.role_desc.capitalize())}</td></tr>"
                f"<tr><td><b><font color='{label_color}'>Value: </font></b></td><td>{self.wrap_and_escape(n.data)}</td></tr>"
            )
            if n.data_id is not None:
                node_label += f"<tr><td><b><font color='{label_color}'>Data ID: </font></b></td><td>{self.wrap_and_escape(n.data_id)}</td></tr>"
            if n.proposing:
                node_label += f"<tr><td><b><font color='{label_color}'>Proposing</font></b></td><td>{{'Yes'}}</td></tr>"
                node_label += f"<tr><td><b><font color='{label_color}'>Previous Value: </font></b></td><td>{self.wrap_and_escape(n.previous_data)}</td></tr>"
            if n.requires_opt:
                node_label += f"<tr><td><b><font color='{label_color}'>Requires Optimization: </font ></b></td><td>{{'Yes'}}</td></tr>"
            if n.param_type:
                node_label += f"<tr><td><b><font color='{label_color}'>Type: </font></b></td><td>{self.wrap_and_escape(n.param_type.name)}</td></tr>"
            if (
                full_trace
                and hasattr(n, "component_trace")
                and n.component_trace.api_kwargs is not None
            ):
                node_label += f"<tr><td><b><font color='{label_color}'> API kwargs: </font></b></td><td>{self.wrap_and_escape(str(n.component_trace.api_kwargs))}</td></tr>"

            # show the score for intermediate nodes
            if n.score is not None and len(n.predecessors) > 0:
                node_label += f"<tr><td><b><font color='{label_color}'>Score: </font></b></td><td>{str(n.score)}</td></tr>"
            if add_grads:
                node_label += f"<tr><td><b><font color='{label_color}'>Gradients: </font></b></td><td>{self.wrap_and_escape(n.get_gradients_names())}</td></tr>"
                # add a list of each gradient with short value
                # combine the gradients and context
                # combined_gradients_contexts = zip(
                #     n.gradients, [n.gradients_context[g] for g in n.gradients]
                # )
                # if "output" in n.name:
                for g in n.gradients:
                    gradient_context = g.context
                    log.info(f"Gradient context display: {gradient_context}")
                    log.info(f"data: {g.data}")
                    node_label += f"<tr><td><b><font color='{label_color}'>Gradient {g.name} Feedback: </font></b></td><td>{self.wrap_and_escape(g.data)}</td></tr>"
                    # if gradient_context != "":
                    #     node_label += f"<tr><td><b><font color='{label_color}'>Gradient {g.name} Context: </font></b></td><td>{wrap_and_escape(gradient_context)}</td></tr>"
                    # if g.prompt:
                    #     node_label += f"<tr><td><b><font color='{label_color}'>Gradient {g.name} Prompt: </font></b></td><td>{wrap_and_escape(g.prompt)}</td></tr>"
            if len(n._traces.values()) > 0:
                node_label += f"<tr><td><b><font color='{label_color}'>Traces: keys: </font></b></td><td>{self.wrap_and_escape(str(n._traces.keys()))}</td></tr>"
                node_label += f"<tr><td><b><font color='{label_color}'>Traces: values: </font></b></td><td>{self.wrap_and_escape(str(n._traces.values()))}</td></tr>"
            if n.tgd_optimizer_trace is not None:
                node_label += f"<tr><td><b><font color='{label_color}'>TGD Optimizer Trace: </font></b></td><td>{self.wrap_and_escape(str(n.tgd_optimizer_trace))}</td></tr>"

            # show component trace, id and name
            if hasattr(n, "component_trace") and n.component_trace.id is not None:
                node_label += f"<tr><td><b><font color='{label_color}'>Component Trace ID: </font></b></td><td>{self.wrap_and_escape(str(n.component_trace.id))}</td></tr>"
            if hasattr(n, "component_trace") and n.component_trace.name is not None:
                node_label += f"<tr><td><b><font color='{label_color}'>Component Trace Name: </font></b></td><td>{self.wrap_and_escape(str(n.component_trace.name))}</td></tr>"

            node_label += "</table>"
            # check if the name exists in dot
            if n.name in node_names:
                n.name = f"{n.name}_{n.id}"
            node_names.add(n.name)
            dot.node(
                name=n.name,
                label=f"<{node_label}>",
                shape="plaintext",
            )
            # writer.add_text(n.name, str(n.to_dict()))
            log.info(f"Node: {n.name}, {n.to_dict()}")
            # track gradients
            for g in n.gradients:

                log.info(f"Gradient: {g.name}, {g.to_dict()}")
                log.info(f"Gradient prompt: {g.prompt}")
        for n1, n2 in edges:
            dot.edge(n1.name, n2.name)

        # dot.render(filepath, format=format, cleanup=True)

        save_json(self.to_dict(), f"{filepath}_root.json")

        # draw interactive graph
        graph_file: Dict[str, str] = self.draw_interactive_html_graph(
            filepath=filepath, nodes=nodes, edges=edges
        )
        output = {
            # "graph_path": final_path,
            "root_path": f"{filepath}_root.json",
            "interactive_html_graph": graph_file["graph_path"],
        }
        print(f"Graph saved as {filepath}.{format}")
        return output

    def draw_output_subgraph(
        self,
        add_grads: bool = True,
        format: str = "png",
        rankdir: str = "TB",
        filepath: str = None,
    ) -> Dict:
        """
        Build and visualize a subgraph containing only OUTPUT parameters.

        Args:
            add_grads (bool): Whether to include gradient edges.
            format (str): Format for output (e.g., png, svg).
            rankdir (str): Graph layout direction ("LR" or "TB").
            filepath (str): Path to save the graph.
        """

        assert rankdir in ["LR", "TB"]
        from adalflow.utils.global_config import get_adalflow_default_root_path

        try:
            from graphviz import Digraph

        except ImportError as e:
            raise ImportError(
                "Please install graphviz using 'pip install graphviz' to use this feature"
            ) from e

        root_path = get_adalflow_default_root_path()

        filename = f"trace_component_output_graph_{self.name}_id_{self.id}.{format}"
        filepath = (
            os.path.join(filepath, filename)
            if filepath
            else os.path.join(root_path, "graphs", filename)
        )

        # Step 1: Collect OUTPUT nodes and edges
        nodes, edges = self._collect_output_subgraph()

        # Step 2: Render using Graphviz
        print(f"Saving OUTPUT subgraph to {filepath}")

        dot = Digraph(format=format, graph_attr={"rankdir": rankdir})
        node_ids = set()

        for node in nodes:

            node_label = f"""
            <table border="0" cellborder="1" cellspacing="0">
                <tr><td><b>Name:</b></td><td>{self.wrap_and_escape(node.name)}</td></tr>
                <tr><td><b>Type:</b></td><td>{self.wrap_and_escape(node.param_type.name)}</td></tr>
                <tr><td><b>Value:</b></td><td>{self.wrap_and_escape(node.get_short_value())}</td></tr>"""
            # add the component trace id and name
            if hasattr(node, "component_trace") and node.component_trace.id is not None:
                escaped_ct_id = html.escape(str(node.component_trace.id))
                node_label += f"<tr><td><b>Component Trace ID:</b></td><td>{escaped_ct_id}</td></tr>"
            if (
                hasattr(node, "component_trace")
                and node.component_trace.name is not None
            ):
                escaped_ct_name = html.escape(str(node.component_trace.name))
                node_label += f"<tr><td><b>Component Trace Name:</b></td><td>{escaped_ct_name}</td></tr>"

            node_label += "</table>"
            dot.node(
                name=node.id,
                label=f"<{node_label}>",
                shape="plaintext",
                color="lightblue" if node.requires_opt else "gray",
            )
            node_ids.add(node.id)

        for source, target in edges:
            if source.id in node_ids and target.id in node_ids:
                dot.edge(source.id, target.id)

        # Step 3: Save and render
        dot.render(filepath, cleanup=True)
        print(f"Graph saved as {filepath}")
        return {"output_subgraph": filepath}

    def draw_component_subgraph(
        self,
        format: str = "png",
        rankdir: str = "TB",
        filepath: str = None,
    ):
        """
        Build and visualize a subgraph containing only OUTPUT parameters.

        Args:
            format (str): Format for output (e.g., png, svg).
            rankdir (str): Graph layout direction ("LR" or "TB").
            filepath (str): Path to save the graph.
        """
        assert rankdir in ["LR", "TB"]
        from adalflow.utils.global_config import get_adalflow_default_root_path

        try:
            from graphviz import Digraph
        except ImportError as e:
            raise ImportError(
                "Please install graphviz using 'pip install graphviz' to use this feature"
            ) from e

        # Step 1: Collect OUTPUT nodes and edges
        component_nodes, edges, component_nodes_orders = (
            self._collect_component_subgraph()
        )
        root_path = get_adalflow_default_root_path()

        # Step 2: Setup graph rendering
        filename = f"output_component_{self.name}_{self.id}.{format}"
        filepath = filepath or f"./{filename}"

        filepath = (
            os.path.join(filepath, filename)
            if filepath
            else os.path.join(root_path, "graphs", filename)
        )
        print(f"Saving OUTPUT subgraph to {filepath}")

        dot = Digraph(format=format, graph_attr={"rankdir": rankdir})

        # Add nodes
        for node in component_nodes:
            node_label = """
            <table border="0" cellborder="1" cellspacing="0">"""

            if node.name:
                node_label += """<tr><td><b>Name:</b></td><td>{node.name}</td></tr>"""
            if node.type:
                node_label += """<tr><td><b>TYPE:</b></td><td>{node.type}</td></tr>"""

            # add the list of orders
            if node.id in component_nodes_orders:
                node_label += f"<tr><td><b>Order:</b></td><td>{component_nodes_orders[node.id]}</td></tr>"
            node_label += "</table>"
            dot.node(
                name=node.id if node.id else "id missing",
                label=f"<{node_label}>",
                shape="plaintext",
                color="lightblue",
            )

        # Add edges with order labels
        for source_id, target_id, edge_order in edges:
            dot.edge(source_id, target_id)  # , label=str(edge_order), color="black")

        # Step 3: Save and render
        dot.render(filepath, cleanup=True)
        print(f"Graph saved as {filepath}")
        return {"component_graph": f"{filepath}"}

    def _collect_output_subgraph(
        self,
    ) -> Tuple[Set["Parameter"], List[Tuple["Parameter", "Parameter"]]]:
        """
        Collect nodes of type OUTPUT and their relationships.

        Returns:
            nodes (Set[Parameter]): Set of OUTPUT nodes.
            edges (List[Tuple[Parameter, Parameter]]): Edges between OUTPUT nodes.
        """
        output_nodes = set()
        edges = []

        visited = set()  # check component_trace.id and name

        def traverse(node: "Parameter"):
            if node in visited:
                return
            visited.add(node)

            # Add OUTPUT nodes to the set
            if (
                node.param_type == ParameterType.OUTPUT
                or "OUTPUT" in node.param_type.name
            ):
                output_nodes.add(node)

            # Traverse predecessors and add edges
            for pred in node.predecessors:
                if (
                    pred.param_type == ParameterType.OUTPUT
                    or "OUTPUT" in pred.param_type.name
                ):
                    edges.append((pred, node))
                traverse(pred)

        traverse(self)
        return output_nodes, edges

    def _collect_component_subgraph(
        self,
    ) -> Tuple[Set[ComponentNode], List[Tuple[str, str]]]:
        """
        Collect OUTPUT nodes and their relationships as ComponentNodes.

        Returns:
            component_nodes (Set[ComponentNode]): Set of component nodes (id and name only).
            edges (List[Tuple[str, str]]): Edges between component IDs.
        """
        component_nodes = set()  # To store component nodes as ComponentNode
        component_nodes_orders: Dict[str, List[int]] = (
            {}
        )  # To store component nodes order
        edges = []  # To store edges between component IDs

        visited = set()  # Track visited parameters to avoid cycles
        edge_counter = [0]  # Mutable counter for edge order tracking

        def traverse(node: "Parameter"):
            if node in visited:
                return
            visited.add(node)

            # Check if node is of OUTPUT type
            if (
                node.param_type == ParameterType.OUTPUT
                or "OUTPUT" in node.param_type.name
            ):
                component_id = node.component_trace.id or f"unknown_id_{uuid.uuid4()}"
                component_name = node.component_trace.name or "Unknown Component"

                # Create a ComponentNode and add to the set
                component_node = ComponentNode(id=component_id, name=component_name)
                component_nodes.add(component_node)

                # Traverse predecessors and add edges
                for pred in node.predecessors:
                    # if pred.param_type != ParameterType.OUTPUT:
                    #     continue
                    pred_id = f"unknown_id_{uuid.uuid4()}"
                    pred_name = "Unknown Component"

                    if hasattr(pred, "component_trace") and pred.component_trace.id:
                        pred_id = pred.component_trace.id
                        pred_name = pred.component_trace.name

                    # Add edge if predecessor is also of OUTPUT type
                    if (
                        pred.param_type == ParameterType.OUTPUT
                        or "OUTPUT" in pred.param_type.name
                    ):
                        edges.append((pred_id, component_id, edge_counter[0]))
                        component_nodes.add(ComponentNode(id=pred_id, name=pred_name))
                        edge_counter[0] += 1

                    if pred.param_type == ParameterType.INPUT:
                        pred_id = pred.id
                        pred_name = pred.name
                        pred_node = ComponentNode(
                            id=pred_id, name=pred_name, type="INPUT"
                        )
                        component_nodes.add(pred_node)
                        # add an edge from input to the first output
                        edges.append((pred_id, component_id, edge_counter[0]))
                        edge_counter[0] += 1

                    traverse(pred)

        # Start traversal from the current parameter
        traverse(self)
        # Reverse the edge order
        # total_edges = len(edges)
        # edges = [
        #     (source, target, (total_edges - 1) - edge_number)
        #     for idx, (source, target, edge_number) in enumerate(edges)
        # ]

        return component_nodes, edges, component_nodes_orders

    def to_dict(self):
        return {
            "name": self.name,
            "id": self.id,
            "role_desc": self.role_desc,
            "data": str(self.data),
            "requires_opt": self.requires_opt,
            "param_type": str(self.param_type),
            # others
            "predecessors": [pred.to_dict() for pred in self.predecessors],
            "gradients": [grad.to_dict() for grad in self.gradients],
            "previous_data": self.previous_data,
            "grad_fn": str(
                self.grad_fn
            ),  # Simplify for serialization, modify as needed
            "score": self.score,
            "traces": {k: v.to_dict() for k, v in self._traces.items()},
            # demos
            "demos": [d.to_dict() for d in self._demos],
        }

    @classmethod
    def from_dict(cls, data: dict):
        predecessors = [cls.from_dict(pred) for pred in data["predecessors"]]
        param = cls(
            name=data["name"],
            role_desc=data["role_desc"],
            data=data["data"],
            requires_opt=data["requires_opt"],
            param_type=ParameterType(data["param_type"]),
            # others
            predecessors=predecessors,
            gradients=[cls.from_dict(grad) for grad in data["gradients"]],
            previous_data=data["previous_data"],
            score=data["score"],
            # demos
            demos=[DataClass.from_dict(d) for d in data["demos"]],
        )
        # Reconstruct gradients_context from the list of tuples
        param._traces = {k: DataClass.from_dict(v) for k, v in data["traces"].items()}
        return param

    # TODO: very hard to read directly, need to simplify and let users use to_dict for better readability
    def __repr__(self):
        return f"Parameter(name={self.name}, requires_opt={self.requires_opt}, param_type={self.param_type}, role_desc={self.role_desc}, data={self.data}, predecessors={self.predecessors}, gradients={self.gradients},\
            traces={self._traces})"


# TODO: separate the Parameter class into different classes and each class will have its own methods instead of all in one class
# class InputParameter(Parameter):
#     """One of the simplest types of parameters, representing an input to the system.
#     Input parameter will not be trainable, but serves a tracing purpose in the computation graph.
#     """

#     def __init__(
#         self,
#         name: str,
#         role_desc: str,
#         data: Any,
#         requires_opt: bool = False,
#         param_type: ParameterType = ParameterType.INPUT,
#     ):
#         super().__init__(
#             name=name,
#             role_desc=role_desc,
#             data=data,
#             requires_opt=requires_opt,
#             param_type=param_type,
#         )


# class HyperParameter(Parameter):
#     """One of the simplest types of parameters, representing a hyperparameter to the system."""

#     def __init__(
#         self,
#         name: str,
#         role_desc: str,
#         data: Any,
#         requires_opt: bool = False,
#         param_type: ParameterType = ParameterType.HYPERPARAM,
#     ):
#         super().__init__(
#             name=name,
#             role_desc=role_desc,
#             data=data,
#             requires_opt=requires_opt,
#             param_type=param_type,
#         )


# class PromptParameter(Parameter):

#     def __init__(
#         self,
#         name: str,
#         role_desc: str,
#         data: Any,
#         requires_opt: bool = True,
#         param_type: ParameterType = ParameterType.PROMPT,
#     ):
#         super().__init__(
#             name=name,
#             role_desc=role_desc,
#             data=data,
#             requires_opt=requires_opt,
#             param_type=param_type,
#         )


# class DemoParameter(Parameter):

#     def __init__(
#         self,
#         name: str,
#         role_desc: str,
#         data: Any,
#         requires_opt: bool = True,
#         param_type: ParameterType = ParameterType.DEMOS,
#     ):
#         super().__init__(
#             name=name,
#             role_desc=role_desc,
#             data=data,
#             requires_opt=requires_opt,
#             param_type=param_type,
#         )


class OutputParameter(Parameter):
    __doc__ = r"""The output parameter is the most complex type of parameter in the system.

    It will trace the predecessors, set up a grad_fn, store gradients, and trace the forward pass by tracking the component_trace.
    """
    allowed_types = {
        ParameterType.OUTPUT,
        ParameterType.LOSS_OUTPUT,
        ParameterType.GENERATOR_OUTPUT,
        ParameterType.SUM_OUTPUT,
    }
    component_trace: ComponentTrace = (
        None  # Trace of the component that produced this output
    )
    full_response: object = None  # The full response from the component

    def __init__(
        self,
        *,
        id: Optional[str] = None,  # unique id of the parameter
        data: T = None,  # for generator output, the data will be set up as raw_response
        data_id: str = None,  # for tracing the data item in the training/val/test set
        requires_opt: bool = True,
        role_desc: str = "",
        param_type: ParameterType = ParameterType.OUTPUT,
        name: str = None,  # name is used to refer to the parameter in the prompt, easier to read for humans
        instruction_to_optimizer: str = None,
        instruction_to_backward_engine: str = None,
        score: Optional[float] = None,
        eval_input: object = None,
        successor_map_fn: Optional[Dict[str, Callable]] = None,
        data_in_prompt: Optional[Callable] = None,
        full_response: Optional[Any] = None,
    ):
        super().__init__(
            id=id,
            data=data,
            data_id=data_id,
            requires_opt=requires_opt,
            role_desc=role_desc,
            param_type=param_type,
            name=name,
            instruction_to_optimizer=instruction_to_optimizer,
            instruction_to_backward_engine=instruction_to_backward_engine,
            score=score,
            eval_input=eval_input,
            successor_map_fn=successor_map_fn,
            data_in_prompt=data_in_prompt,
        )

        self.component_trace = ComponentTrace()
        self.full_response = full_response

    ############################################################################################################
    #  Trace component, include trace_forward_pass & trace_api_kwargs for now
    ############################################################################################################
    def trace_forward_pass(
        self,
        input_args: Dict[str, Any],
        full_response: object,
        id: str = None,
        name: str = None,
    ):
        r"""Trace the forward pass of the parameter. Adding the component information to the trace"""
        self.input_args = input_args
        self.full_response = full_response
        # TODO: remove the input_args and full_response to use component_trace
        self.component_trace.input_args = input_args
        self.component_trace.full_response = full_response
        self.component_trace.id = id
        self.component_trace.name = name
        # just for convenience to trace full response separately
        self.full_response = full_response

    def trace_api_kwargs(self, api_kwargs: Dict[str, Any]):
        r"""Trace the api_kwargs for components like Generator and Retriever that pass to the model client."""
        self.component_trace.api_kwargs = api_kwargs

    def to_dict(self):
        super_dict = super().to_dict()
        super_dict.update(
            {
                "component_trace": self.component_trace.to_dict(),
            }
        )

    @classmethod
    def from_dict(cls, data: dict):
        component_trace = ComponentTrace.from_dict(data["component_trace"])
        return super().from_dict(data).update({"component_trace": component_trace})

    def __repr__(self):
        super_repr = super().__repr__()
        start = super_repr.find("Parameter")
        if start == 0:
            end = start + len("Parameter")
            super_repr = super_repr[:start] + "OutputParameter" + super_repr[end:]
        return super_repr


if __name__ == "__main__":

    # test gradient hash and to_dict
    from_response = OutputParameter(
        name="p1",
        role_desc="role1",
        data=1,
    )
    from_response.component_trace = ComponentTrace(id="1")
    g1 = Gradient(
        from_response=from_response,
        to_pred=Parameter(name="p2", role_desc="role2", data=2),
        data_id="1",
    )
    g2 = Gradient(
        from_response=from_response,
        to_pred=Parameter(name="p2", role_desc="role2", data=2),
        data_id="1",
    )
    print(g1 == g2)
    print(g1.__hash__())
    print(g2.__hash__())
    print(isinstance(g1, Gradient))  # Should print True

    print(g1.to_dict())



"""Implementation of G-Eval: G-eval <https://arxiv.org/abs/2303.08774, https://github.com/nlpyang/geval>
Instead of getting 1/5 as the score, AdalFlow will use 0.2 as the score, so that we can have a score in range [0, 1] for all metrics."""

from enum import Enum
from typing import Dict, Any, Optional, List, Tuple
import logging


from adalflow.eval.base import BaseEvaluator
from adalflow.core.component import Component
from adalflow.core.model_client import ModelClient
from adalflow.eval.llm_as_judge import DEFAULT_LLM_EVALUATOR_MODEL_KWARGS
from adalflow.core.string_parser import FloatParser

__all__ = ["GEvalMetric", "NLGTask", "GEvalLLMJudge", "GEvalJudgeEvaluator"]

log = logging.getLogger(__name__)


class GEvalMetric(Enum):

    RELEVANCE = "Relevance"  # range [1, 5]
    FLUENCY = "Fluency"  # range [1, 3]
    CONSISTENCY = "Consistency"  # range [1, 5]
    COHERENCE = "Coherence"  # range [1, 5]


all_geval_metrics = [m.value for m in GEvalMetric]
all_geval_metrics_to_range = {
    "Relevance": 5.0,
    "Fluency": 3.0,
    "Consistency": 5.0,
    "Coherence": 5.0,
}


class NLGTask(Enum):
    SUMMARIZATION = {
        "task_desc_str": r"""You will be given a summary of a text.  Please evaluate the summary based on the following criteria:""",
        "criteria_relevance": r"""Relevance (1-5) - selection of important content from the source.
        The summary should include only important information from the source document.
        Annotators were instructed to penalize summaries which contained redundancies and excess information.""",
        "steps_relevance": r"""1. Read the summary and the source document carefully.
        2. Compare the summary to the source document and identify the main points of the article.
        3. Assess how well the summary covers the main points of the article, and how much irrelevant or redundant information it contains.
        4. Assign a relevance score from 1 to 5.""",
        "criteria_fluency": r"""Fluency (1-3): the quality of the summary in terms of grammar, spelling, punctuation, word choice, and sentence structure.
        - 1: Poor. The summary has many errors that make it hard to understand or sound unnatural.
        - 2: Fair. The summary has some errors that affect the clarity or smoothness of the text, but the main points are still comprehensible.
        - 3: Good. The summary has few or no errors and is easy to read and follow.
        """,
        "steps_fluency": None,
        "criteria_coherence": r"""Coherence (1-5) - the collective quality of all sentences.
        We align this dimension with the DUC quality question of structure and coherence whereby "the summary should be well-structured and well-organized.
        The summary should not just be a heap of related information, but should build from sentence to a coherent body of information about a topic.""",
        "steps_coherence": r"""1. Read the input text carefully and identify the main topic and key points.
        2. Read the summary and assess how well it captures the main topic and key points. And if it presents them in a clear and logical order.
        3. Assign a score for coherence on a scale of 1 to 5, where 1 is the lowest and 5 is the highest based on the Evaluation Criteria.""",
        "criteria_consistency": r"""Consistency (1-5) - the factual alignment between the summary and the summarized source.
        A factually consistent summary contains only statements that are entailed by the source document.
        Annotators were also asked to penalize summaries that contained hallucinated facts. """,
        "steps_consistency": r"""1. Read the summary and the source document carefully.
        2. Identify the main facts and details it presents.
        3. Read the summary and compare it to the source document to identify any inconsistencies or factual errors that are not supported by the source.
        4. Assign a score for consistency based on the Evaluation Criteria.""",
    }


DEFAULT_G_EVAL_RPROMPT = r"""
<START_OF_SYSTEM_PROMPT>
{# task desc #}
{{task_desc_str}}
---------------------
{# evaluation criteria #}
Evaluation Criteria:
{{evaluation_criteria_str}}
---------------------
{# evaluation steps #}
{% if evaluation_steps_str %}
Evaluation Steps:
{{evaluation_steps_str}}
---------------------
{% endif %}
{{input_str}}
{ # evaluation form #}
Output the score for metric (scores ONLY!): {{metric_name}}
<END_OF_SYSTEM_PROMPT>
"""


class GEvalLLMJudge(Component):
    __doc__ = r"""Demonstrate how to use an LLM/Generator to output True or False for a judgement query.

    You can use any of your template to adapt to more tasks and sometimes you can directly ask LLM to output a score in range [0, 1] instead of only True or False.

    A call on the LLM judge equalize to _compute_single_item method.

    Args:
        model_client (ModelClient): The model client to use for the generator.
        model_kwargs (Dict[str, Any], optional): The model kwargs to pass to the model client. Defaults to {}. Please refer to :ref:`ModelClient<components-model_client>` for the details on how to set the model_kwargs for your specific model if it is from our library.
        template (str, optional): The template to use for the LLM evaluator. Defaults to None.
        use_cache (bool, optional): Whether to use cache for the LLM evaluator. Defaults to True.
        default_task (NLGTask, optional): The default task to use for the judgement query. Defaults to None.
   """

    def __init__(
        self,
        model_client: Optional[ModelClient] = None,
        model_kwargs: Optional[Dict[str, Any]] = None,
        template: Optional[str] = None,
        use_cache: bool = True,
        default_task: Optional[NLGTask] = None,
    ):
        from adalflow.core.generator import Generator

        super().__init__()
        self.model_client = model_client
        if model_client is None:
            log.info("model_client is None, default to OpenAIClient.")
            try:
                from adalflow.components.model_client import OpenAIClient
            except ImportError:
                raise ImportError(
                    "OpenAIClient is not available. Please fix the import error or set your own choice of model_client and model_kwargs."
                )
            self.model_client = OpenAIClient()
        self.model_kwargs = model_kwargs or DEFAULT_LLM_EVALUATOR_MODEL_KWARGS
        self.template = template or DEFAULT_G_EVAL_RPROMPT
        self.prompt_kwargs = {k: {} for k in all_geval_metrics}
        self.default_task = default_task
        if default_task:
            # task_name = default_task.name
            # print(f"task_name: {task_name}")
            for metric_name in all_geval_metrics:
                metric_name_lower = metric_name.lower()
                self.prompt_kwargs[metric_name] = {
                    "task_desc_str": default_task.value["task_desc_str"],
                    "evaluation_criteria_str": default_task.value[
                        f"criteria_{metric_name_lower}"
                    ],
                    "evaluation_steps_str": default_task.value.get(
                        f"steps_{metric_name_lower}"
                    ),
                    "metric_name": metric_name,
                }
        self.llm_evaluator = Generator(
            model_client=self.model_client,
            model_kwargs=self.model_kwargs,
            template=self.template,
            use_cache=use_cache,
            output_processors=FloatParser(),
        )

    def call(self, input_str: str) -> Dict[str, Any]:
        r"""
        Pass the input string with all information to the LLM evaluator and get the judgement.

        Args:
            input_str (str): The input string with all information.

        Returns:
            Dict[str, Any]: The judgement result.
        """

        output = {}

        total_scores = 0
        num_metrics = 0

        for metric_name in all_geval_metrics:
            # add input_str to prompt_kwargs
            self.prompt_kwargs[metric_name]["input_str"] = input_str

            metric_score = self.llm_evaluator(
                prompt_kwargs=self.prompt_kwargs[metric_name]
            )
            # print(
            #     f"prompt: \n{self.llm_evaluator.get_prompt(**self.prompt_kwargs[metric_name])}"
            # )
            output[metric_name] = (
                metric_score.data / all_geval_metrics_to_range[metric_name]
                if metric_score and metric_score.data
                else None
            )
            # print(f"metric score: {metric_score}")
            if output[metric_name]:
                total_scores += output[metric_name]
                num_metrics += 1
        output["overall"] = total_scores / num_metrics if num_metrics else None

        return output

    def _extra_repr(self) -> str:
        s = f"default_task= {self.default_task}, prompt_kwargs={self.prompt_kwargs}"
        return s


class GEvalJudgeEvaluator(BaseEvaluator):
    r"""
    LLM as judge for evaluating the performance of a LLM in form of GEval with 4 main metrics:

    Relevance, Fluency, Consistency, Coherence.

    Args:
        llm_judge (Component, optional): The LLM evaluator to use. Defaults to GEvalLLMJudge().
    """

    def __init__(
        self,
        llm_judge: Optional[Component] = None,
    ):
        super().__init__()
        self.llm_judge = llm_judge or GEvalLLMJudge()

    def compute_single_item(self, input_str: str) -> Dict[str, Any]:
        r"""
        Compute the score for a single item.

        Args:
            input_str (str): The input string with all information.

        Returns:
            Dict[str, Any]: The judgement result.
        """
        return self.llm_judge(input_str=input_str)

    def compute(
        self,
        input_strs: List[str],
    ) -> Tuple[Dict, List[Dict[str, Any]]]:
        r"""
        Get the judgement of the predicted answer for a list of questions.

        Args:
           input_strs (List[str]): List of input strings.
        Returns:
            List[Dict[str, Any]]: The judgement result.
        """
        output = []
        for input_str in input_strs:
            output.append(self.compute_single_item(input_str))

        # average across different keys
        final_output = {metric: [] for metric in all_geval_metrics}
        final_output.update({"overall": []})
        for data in output:
            for metric, score in data.items():
                if score is None:
                    continue
                final_output[metric].append(score)

        for metric, scores in final_output.items():
            if not scores:
                final_output[metric] = None
            else:
                final_output[metric] = sum(scores) / len(scores)

        return final_output, output

    def __str__(self) -> str:
        s = f"llm_judge={self.llm_judge}, prompt_kwargs={self.llm_judge.prompt_kwargs}"
        return s



"""LocalDB to perform in-memory storage and data persistence(pickle or any filesystem) for data models like documents and dialogturn."""

from typing import List, Optional, Callable, Dict, Any, TypeVar, Generic, overload
import logging
import os
from dataclasses import field, dataclass
import pickle


from adalflow.core.component import Component
from adalflow.utils.registry import EntityMapping
from adalflow.utils.global_config import get_adalflow_default_root_path


log = logging.getLogger(__name__)

T = TypeVar("T")  # Allow any type as items

U = TypeVar("U")  # U will be the type after transformation


# TODO: localDB does not need to be a component
# TODO: DB clarity can be further improved
@dataclass
class LocalDB(Generic[T], Component):
    __doc__ = """LocalDB with in-memory CRUD operations, data transformation/processing pipelines, and persistence.

    LocalDB is highly flexible.
    1. It can store any type of data items in the `items` attribute.
    2. You can register and apply multiple transformers, and save the transformed data in the `transformed_items` attribute.
       This is highly useful to manage experiments with different data transformations.
    3. You can save the state of the LocalDB to a pickle file and load it back later. All states are restored.
        str(local_db.__dict__) == str(local_db_loaded.__dict__) should be True.

    .. note::
        The transformer should be of type Component. We made the effort in the library to make every component picklable.

    CRUD operations:
    1. Create a new db: ``db = LocalDB(name="my_db")``
    2. load: Load the db with data. ``db.load([{"text": "hello world"}, {"text": "hello world2"}])``
    3. extend: Extend the db with data. ``db.extend([{"text": "hello world3"}])``.
       In default, the transformer is applied and the transformed data is extended.
    4. add: Add a single item to the db. ``db.add({"text": "hello world4"})``.
       In default, the transformer is applied and the transformed data is added.
       Unless the transformed data keeps the same length as the original data, the insert operation does not mean insert after the last item.
    5. delete: Remove items by index. ``db.delete([0])``.
    6. reset: Remove all items. ``db.reset()``, including transformed_items and transformer_setups,and mapper_setups.

    Data transformation:
    1. Register a transformer first and apply it later

    .. code-block:: python

            db.register_transformer(transformer, key="test", map_fn=map_fn)
            # load data
            db.load([{"text": "hello world"}, {"text": "hello world2"}], apply_transformer=True)

            # or load data first and apply transformer by key
            db.load([{"text": "hello world"}, {"text": "hello world2"}], apply_transformer=False)
            db.apply_transformer("test")

    2. Add a version of transformed data to the db along with the transformer.

    .. code-block:: python

            db.transform(transformer, key="test", map_fn=map_fn)

    Data persistence:
    1. Save the state of the db to a pickle file.

    .. code-block:: python

            db.save_state("storage/local_item_db.pkl")

    2. Load the state of the db from a pickle file.

    .. code-block:: python

            db2 = LocalDB.load_state("storage/local_item_db.pkl")

    3. Check if the loaded and original db are the same.

    .. code-block:: python

                str(db.__dict__) == str(db2.__dict__) # expect True

    Args:

        items (List[T], optional): The original data items. Defaults to []. Can be any type such as Document, DialogTurn, dict, text, etc.
            The only requirement is that they should be picklable/serializable.
        transformed_items (Dict[str, List [U]], optional): Transformed data items by key. Defaults to {}.
             Transformer must be of type Component.
        transformer_setups (Dict[str, Component], optional): Transformer setup by key. Defaults to {}.
          It is used to save the transformer setup for later use.
        mapper_setups (Dict[str, Callable[[T], Any]], optional): Map function setup by key. Defaults to {}.
    """

    name: Optional[str] = field(
        default=None, metadata={"description": "Name of the DB"}
    )
    items: List[T] = field(
        default_factory=list, metadata={"description": "The original data items"}
    )

    transformed_items: Dict[str, List[U]] = field(
        default_factory=dict, metadata={"description": "Transformed data items by key"}
    )

    transformer_setups: Dict[str, Component] = field(
        default_factory=dict, metadata={"description": "Transformer setup by key"}
    )
    mapper_setups: Dict[str, Callable[[T], Any]] = field(
        default_factory=dict, metadata={"description": "Map function setup by key"}
    )
    index_path: Optional[str] = field(
        default="index.faiss", metadata={"description": "Path to the index file"}
    )

    def __post_init__(self):
        super().__init__()

    @property
    def length(self):
        return len(self.items)

    def get_transformer_keys(self) -> List[str]:
        return list(self.transformed_items.keys())

    # TODO: combine this to fetch_transformed_items
    def get_transformed_data(
        self, key: str, filter_fn: Callable[[Any], bool] = lambda x: True
    ) -> List[U]:
        """
        Get the transformed items by key after applying a filter on metadata.

        Args:
            key (str): The key to identify which transformed items to retrieve.
            filter_fn (Callable[[Any], bool], optional): The filter function to apply on the metadata. Defaults to lambda x: True.

        Returns:
            List[U]: The filtered and transformed items.
        """
        if key not in self.transformed_items:
            raise ValueError(f"Key {key} not found in transformed items.")
        # Apply filter function on the transformed items
        return list(filter(filter_fn, self.transformed_items[key]))

    def _get_transformer_name(self, transformer: Component) -> str:
        name = f"{transformer.__class__.__name__}_"
        for n, _ in transformer.named_components():
            name += n + "_"
        return name

    def register_transformer(
        self,
        transformer: Component,
        key: Optional[str] = None,
        map_fn: Optional[Callable[[T], Any]] = None,
    ) -> str:
        """Register a transformer to be used later for transforming the data."""
        if key is None:
            key = self._get_transformer_name(transformer)
            log.info(f"Generated key for transformer: {key}")
        self.transformer_setups[key] = transformer
        if map_fn is not None:
            self.mapper_setups[key] = map_fn
        self.transformed_items[key] = []
        return key

    @overload
    def transform(self, key: str) -> str:
        """Apply the transformer by key to the data."""
        ...

    @overload
    def transform(
        self,
        transformer: Component,
        key: Optional[str] = None,
        map_fn: Optional[Callable[[T], Any]] = None,
    ) -> str:
        """Register and apply the transformer to the data."""
        ...

    def transform(
        self,
        transformer: Optional[Component] = None,
        key: Optional[str] = None,
        map_fn: Optional[Callable[[T], Any]] = None,
    ) -> str:
        """The main method to apply the transformer to the data in two ways:
        1. Apply the transformer by key to the data using ``transform(key="test")``.
        2. Register and apply the transformer to the data using ``transform(transformer, key="test")``.

        Args:
            transformer (Optional[Component], optional): The transformer to use. Defaults to None.
            key (Optional[str], optional): The key to use for the transformer. Defaults to None.
            map_fn (Optional[Callable[[T], Any]], optional): The map function to use. Defaults to None.

        Returns:
            str: The key used for the transformation, from which the transformed data can be accessed.
        """
        key_to_use = key
        if transformer:
            key = self.register_transformer(transformer, key, map_fn)
            key_to_use = key
        if key_to_use is None:
            raise ValueError("Key must be provided.")

        if map_fn is not None:
            items_to_use = [map_fn(item) for item in self.items]
        else:
            items_to_use = self.items.copy()

        transformer_to_use = self.transformer_setups[key_to_use]
        self.transformed_items[key_to_use] = transformer_to_use(items_to_use)
        return key_to_use

    def load(self, items: List[Any]):
        """Load the db with new items.

        Args:
            items (List[Any]): The items to load.

        Examples:

        .. code-block:: python

            db = LocalDB()
            db.load([{"text": "hello world"}, {"text": "hello world2"}])
        """
        self.items = items

    def extend(
        self,
        items: List[Any],
        apply_transformer: bool = True,
    ):
        """Extend the db with new items."""

        self.items.extend(items)

        if apply_transformer:
            for key, transformer in self.transformer_setups.items():
                # check if there was a map function registered
                transformed_items = []
                if key in self.mapper_setups:
                    map_fn = self.mapper_setups[key]
                    transformed_items = transformer([map_fn(doc) for doc in items])
                else:
                    transformed_items = transformer(items)
                self.transformed_items[key].extend(transformed_items)

    def delete(self, index: Optional[int] = None, remove_transformed: bool = True):
        """Remove items by index or pop the last item. Optionally remove the transformed data as well.

        Assume the transformed item has the same index as the original item. Might not always be the case.

        Args:
            index (Optional[int], optional): The index to remove. Defaults to None.
            remove_transformed (bool, optional): Whether to remove the transformed data as well. Defaults to True.
        """
        if remove_transformed:
            for key in self.transformed_items.keys():
                self.transformed_items[key].pop(index)
        self.items.pop(index)

    def add(
        self, item: Any, index: Optional[int] = None, apply_transformer: bool = True
    ):
        """Add a single item by index or append to the end. Optionally apply the transformer.

        .. note::
            The item will be transformed using the registered transformer.
            Only if the transformed data keeps the same length as the original data, the ``insert`` operation will work correctly.

        Args:
            item (Any): The item to add.
            index (int, optional): The index to add the item at. Defaults to None.
            When None, the item is appended to the end.
            apply_transformer (bool, optional): Whether to apply the transformer to the item. Defaults to True.
        """
        transformed_items: Dict[str, List] = {}
        if apply_transformer:
            for key, transformer in self.transformer_setups.items():
                transformed_docs = []
                map_fn = self.mapper_setups.get(key, None)
                if map_fn is not None:
                    transformed_docs = transformer([map_fn(item)])
                else:
                    transformed_docs = transformer([item])
                transformed_items[key] = transformed_docs

        if index is not None:
            self.items.insert(index, item)
            for key, transformed_docs in transformed_items.items():
                for doc in transformed_docs:
                    self.transformed_items[key].insert(index, doc)
        else:
            self.items.append(item)
            for key, transformed_docs in transformed_items.items():
                self.transformed_items[key].extend(transformed_docs)

    # TODO: rename it better to add the condition filter
    def fetch_items(self, condition: Callable[[T], bool]) -> List[T]:
        """Fetch items with a condition."""
        return [item for item in self.items if condition(item)]

    def fetch_transformed_items(
        self, key: str, condition: Callable[[U], bool]
    ) -> List[U]:
        """Fetch transformed items with a condition."""
        return [item for item in self.transformed_items[key] if condition(item)]

    def reset(self):
        r"""Reset all attributes to empty."""
        self.mapped_items = {}
        self.transformer_setups = {}
        self.mapper_setups = {}
        self.items = []

    def save_state(self, filepath: str = None):
        """Save the current state (attributes) of the DB using pickle.

        Note:
            The transformer setups will be lost when pickling. As it might not be picklable.
        """
        filepath = filepath or os.path.join(
            get_adalflow_default_root_path,
            (
                "local_db/local_item_db.pkl"
                if not self.name
                else f"local_db/{self.name}.pkl"
            ),
        )
        self.index_path = filepath
        file_dir = os.path.dirname(filepath)
        if not os.path.exists(file_dir):
            os.makedirs(file_dir, exist_ok=True)

        with open(filepath, "wb") as file:
            pickle.dump(self, file)
        print(f"Saved the state of the DB to {filepath}")

    @classmethod
    def load_state(cls, filepath: str = None) -> "LocalDB":
        """Load the state of the DB from a pickle file."""
        filepath = filepath or os.path.join(
            get_adalflow_default_root_path, "local_db/local_item_db.pkl"
        )
        if os.path.exists(filepath):
            with open(filepath, "rb") as file:
                return pickle.load(file)

    def __getstate__(self):
        """Special handling of the components in pickling."""
        state = self.__dict__.copy()
        _transformer_files = {}
        _transformer_type_names = {}

        for key, transformer in self.transformer_setups.items():
            _transformer_files[key] = transformer.to_dict()
            _transformer_type_names[key] = transformer.__class__.__name__
        state["transformer_setups"] = {}
        state["_transformer_files"] = _transformer_files
        state["_transformer_type_names"] = _transformer_type_names
        return state

    def __setstate__(self, state):
        """Restore state with special handling of the components."""
        _transformer_files = state.pop("_transformer_files")
        _transformer_type_names = state.pop("_transformer_type_names")
        self.__dict__.update(state)
        for key, transformer_file in _transformer_files.items():
            class_type = (
                EntityMapping.get(_transformer_type_names[key])
                or globals()[_transformer_type_names[key]]
            )
            self.transformer_setups[key] = class_type.from_dict(transformer_file)
